"""
Bridge module to use the SkyCoord API for SPICE computations.

.. note::
    This module requires the optional dependency `~spiceypy.spiceypy` to be
    installed.

The `SPICE <https://naif.jpl.nasa.gov/naif/>`__ observation geometry information
system is being increasingly used by space missions to describe the locations of
spacecraft and the time-varying orientations of reference frames.
While the `~spiceypy.spiceypy` package provides a Python interface for
performing SPICE computations, its API is very different from that of
`~astropy.coordinates.SkyCoord`.

This module "wraps" `~spiceypy.spiceypy` functionality so that relevant SPICE
computations can be accessed using the `~astropy.coordinates.SkyCoord` API.
When loading a set of kernels, a frame class and corresponding transformations
are created for each SPICE frame. One can also query the location of a body
as computed via SPICE or retrieve the field of view (FOV) of an instrument.

To facilitate the conversion of a SPICE-based coordinate to the built-in frames
in `sunpy.coordinates`, every SPICE-based coordinate has the method
:meth:`~sunpy.coordinates.spice.SpiceBaseCoordinateFrame.to_helioprojective()`.
This method returns a coordinate in the `~sunpy.coordinates.Helioprojective`
frame with the ``observer`` frame attribute appropriately set.

Be aware that when converting a SPICE-based coordinate to/from a built-in frame,
there can be small inconsistencies due to differing planetary ephemerides and
models for various orientations.

Notes
-----
* 2D coordinates can be transformed only if the to/from frames have the same
  SPICE body ID as their centers.
* Transformations of velocities are not yet supported.
* SPICE frame names are rendered as uppercase, except for plus/minus characters,
  which are replaced with lowercase ``'p'``/``'n'`` characters.

Examples
--------
.. minigallery:: sunpy.coordinates.spice.initialize
"""
# Developer notes:
# * We create a public SkyCoord frame for each SPICE frame that is defined in
#   the kernels, but this does not include built-in SPICE frames (e.g., inertial
#   frames or IAU_* frames). The user needs to manually install each built-in
#   SPICE frame that they want to use because otherwise there are simply too
#   many.
# * We also create a private SkyCoord frame for each unique SPICE frame center.
#   Each SPICE frame defines its center, and typically many frames share the
#   same center. By creating these private frames for frame centers, we can
#   transform 2D coordinates between frames that share the same center because
#   the origin does not change.
# * Any transformation that involves a change in frame center (including even
#   a change in the body ID that still maps to the same location) will be
#   treated as a change in origin, and the transformation is routed through
#   ICRS. ICRS is a safe frame to use because the SPICE built-in inertial
#   frame 'J2000' is ICRS, despite its name.

import numpy as np

try:
    import spiceypy
except ImportError:
    raise ImportError("This module requires the optional dependency `spiceypy`.")

import astropy.units as u
from astropy.coordinates import ICRS, ConvertError, SkyCoord, frame_transform_graph
from astropy.coordinates.matrix_utilities import rotation_matrix
from astropy.coordinates.representation import CartesianRepresentation, SphericalRepresentation
from astropy.coordinates.transformations import FunctionTransformWithFiniteDifference
from astropy.time import Time

from sunpy import log
from sunpy.coordinates import SunPyBaseCoordinateFrame
from sunpy.time import parse_time
from sunpy.time.time import _variables_for_parse_time_docstring
from sunpy.util.decorators import add_common_docstring

__all__ = ['SpiceBaseCoordinateFrame', 'get_body', 'get_fov', 'initialize', 'install_frame', 'get_rotation_matrix']


# Note that this epoch is very slightly different from the typical definition of J2000.0 (in TT)
_ET_REF_EPOCH = Time('J2000', scale='tdb')

_CLASS_TYPES = {1: 'inertial', 2: 'PCK', 3: 'CK', 4: 'TK', 5: 'dynamic', 6: 'switch'}


# Registry of the generated frame classes and center classes
_frame_registry = {}
_center_registry = {'SOLAR SYSTEM BARYCENTER': ICRS}


@add_common_docstring(**_variables_for_parse_time_docstring())
class SpiceBaseCoordinateFrame(SunPyBaseCoordinateFrame):
    """
    Base class for all frames generated to represent SPICE frames.

    This class is not intended to be used directly and has no transformations
    defined.

    Parameters
    ----------
    obstime : {parse_time_types}
        The time of the observation. This is used to determine the
        position of solar-system bodies (e.g., the Sun and the Earth) as
        needed to define the origin and orientation of the frame.
    """
    def __init_subclass__(cls, **kwargs):
        cls._frame_name = kwargs.pop('frame_name', None)
        cls._center_name = kwargs.pop('center_name', None)

        super().__init_subclass__(**kwargs)

        cls.__doc__ = (f"Coordinate frame for the SPICE frame '{cls._frame_name}'\n\n"
                       f"Origin: '{cls._center_name}'\n\n"
                       "Parameters\n----------\n"
                       f"obstime : {_variables_for_parse_time_docstring()['parse_time_types']}\n"
                       "    The time of the observation. This is used to determine the\n"
                       "    position of solar-system bodies (e.g., the Sun and the Earth) as\n"
                       "    needed to define the origin and orientation of the frame.\n")

    def to_helioprojective(self):
        """
        Convert this coordinate to the Helioprojective frame.

        The center of the frame center is used as the ``observer`` of the
        `~sunpy.coordinates.Helioprojective` frame.

        Examples
        --------
        .. minigallery:: sunpy.coordinates.spice.SpiceBaseCoordinate.to_helioprojective
        """
        et = _convert_to_et(self.obstime)

        # Get the matrix to rotate from the SPICE frame to heliographic coordinates
        # matrix needs to be contiguous (see https://github.com/astropy/astropy/issues/15503)
        frame_to_iau = np.ascontiguousarray(spiceypy.sxform(self._frame_name, 'IAU_SUN', et)[..., :3, :3])

        # Get the observer location in heliographic coordinates
        obs_iau = spiceypy.spkpos(self._center_name, et, 'IAU_SUN', 'NONE', 'SUN')[0] << u.km
        obs_iau = CartesianRepresentation(obs_iau.T).represent_as(SphericalRepresentation)

        # Construct the matrix to rotate from heliographic coordinates to HPC-like coordinates
        iau_to_hpc = rotation_matrix(-obs_iau.lat, axis='y') @ rotation_matrix(obs_iau.lon, axis='z')

        # To get to actual HPC coordinates, need to flip the X axis
        flip_x = np.array([[-1, 0, 0], [0, 1, 0], [0, 0, 1]])

        # Transform the data by all of the sequence of these matrices to get the HPC vector
        hpc_repr = self.cartesian.transform(flip_x @ iau_to_hpc @ frame_to_iau)

        # Get the observer location in ICRS, with obstime define to be able to transform to HGS
        obs_icrs = spiceypy.spkpos(self._center_name, et, 'J2000', 'NONE', 'SSB')[0] << u.km
        obs_sc = SkyCoord(CartesianRepresentation(obs_icrs.T), frame='icrs', obstime=self.obstime)

        # Construct the HPC coordinate from the vector and the observer
        out_sc = SkyCoord(hpc_repr, frame='helioprojective', obstime=self.obstime, observer=obs_sc)
        if _is_2d(hpc_repr):
            out_sc.representation_type = 'unitspherical'
        return out_sc


def _convert_to_et(time):
    return (time - _ET_REF_EPOCH).to_value('s')


def _astropy_frame_name(spice_frame_name):
    # Replace plus/minus characters in the SPICE frame name with lowercase 'p'/'n'
    return f"spice_{spice_frame_name.translate(str.maketrans('+-', 'pn'))}"


def _astropy_center_name(spice_center_id):
    # Use the center ID directly, changing a negative sign to 'n'
    return f"_spice_center_{str(spice_center_id).replace('-', 'n')}"


def _is_2d(data):
    return data.norm().unit is u.one and u.allclose(data.norm(), 1*u.one)


def _install_center_by_id(center_id):
    center_name = spiceypy.bodc2n(center_id)
    if center_name in _center_registry.keys():
        return

    log.info(f"Creating ICRF frame with {center_name} ({center_id}) origin")
    astropy_center_name = _astropy_center_name(center_id)
    center_cls = type(astropy_center_name, (SpiceBaseCoordinateFrame,), {},
                      frame_name=None, center_name=center_name)

    @frame_transform_graph.transform(FunctionTransformWithFiniteDifference, ICRS, center_cls)
    def icrs_to_shifted(from_icrs_coord, to_shifted_frame):
        if _is_2d(from_icrs_coord.data):
            raise ConvertError("Cannot transform a 2D coordinate due to a shift in origin.")
        icrs_offset = spiceypy.spkpos(center_name,
                                      _convert_to_et(to_shifted_frame.obstime),
                                      'J2000', 'NONE', 'SSB')[0] << u.km
        shifted_pos = from_icrs_coord.cartesian - CartesianRepresentation(icrs_offset.T)
        return to_shifted_frame.realize_frame(shifted_pos)

    @frame_transform_graph.transform(FunctionTransformWithFiniteDifference, center_cls, ICRS)
    def shifted_to_icrs(from_shifted_coord, to_icrs_frame):
        if _is_2d(from_shifted_coord.data):
            raise ConvertError("Cannot transform a 2D coordinate due to a shift in origin.")
        icrs_offset = spiceypy.spkpos(center_name,
                                      _convert_to_et(from_shifted_coord.obstime),
                                      'J2000', 'NONE', 'SSB')[0] << u.km
        icrs_pos = from_shifted_coord.cartesian + CartesianRepresentation(icrs_offset.T)
        return to_icrs_frame.realize_frame(icrs_pos)

    frame_transform_graph._add_merged_transform(center_cls, ICRS, center_cls)

    _center_registry[center_name] = center_cls


def _install_frame_by_id(frame_id):
    frame_name = spiceypy.frmnam(frame_id)
    astropy_frame_name = _astropy_frame_name(frame_name)

    center_id, class_num, _ = spiceypy.frinfo(frame_id)
    center_name = spiceypy.bodc2n(center_id)
    log.info(f"Installing {frame_name} {_CLASS_TYPES[class_num]} frame ({frame_id}) "
             f"as '{astropy_frame_name}'")

    # Create a center class (if needed)
    _install_center_by_id(center_id)
    center_cls = _center_registry[center_name]

    frame_cls = type(astropy_frame_name, (SpiceBaseCoordinateFrame,), {},
                     frame_name=frame_name, center_name=center_name)
    # Force the capitalization pattern of lowercase "spice_" followed by uppercase SPICE frame name
    frame_cls.name = frame_cls.__name__

    @frame_transform_graph.transform(FunctionTransformWithFiniteDifference, center_cls, frame_cls)
    def rotate_from_icrf(from_shifted_coord, to_spice_frame):
        et = _convert_to_et(to_spice_frame.obstime)
        # matrix needs to be contiguous (see https://github.com/astropy/astropy/issues/15503)
        matrix = np.ascontiguousarray(spiceypy.sxform('J2000', frame_name, et)[..., :3, :3])
        new_pos = from_shifted_coord.data.transform(matrix)
        return to_spice_frame.realize_frame(new_pos)

    @frame_transform_graph.transform(FunctionTransformWithFiniteDifference, frame_cls, center_cls)
    def rotate_to_icrf(from_spice_coord, to_shifted_frame):
        et = _convert_to_et(from_spice_coord.obstime)
        # matrix needs to be contiguous (see https://github.com/astropy/astropy/issues/15503)
        matrix = np.ascontiguousarray(spiceypy.sxform(frame_name, 'J2000', et)[..., :3, :3])
        shifted_pos = from_spice_coord.data.transform(matrix)
        return to_shifted_frame.realize_frame(shifted_pos)

    frame_transform_graph._add_merged_transform(frame_cls, center_cls, frame_cls)

    _frame_registry[frame_name] = (frame_cls, center_cls)


def _uninstall_frame_by_class(target_class, from_class):
    frame_transform_graph.remove_transform(target_class, target_class, None)
    frame_transform_graph.remove_transform(from_class, target_class, None)
    frame_transform_graph.remove_transform(target_class, from_class, None)
    del target_class


def initialize(kernels):
    """
    Load one more more SPICE kernels and create corresponding frame classes.

    Parameters
    ----------
    kernels : `str`, `list` of `str`
         One or more SPICE kernel files

    Notes
    -----
    If a kernel file is a meta-kernel, make sure that the relative paths therein
    are correct for the current working directory, which may not be the same as the
    location of the meta-kernel file.

    This function has simple support for being called multiple times in a session in
    order to load multiple sets of kernels. However, there may be unexpected
    behavior if this function is called after the frame classes start being used.

    Examples
    --------
    .. minigallery:: sunpy.coordinates.spice.initialize
    """
    if not isinstance(kernels, list):
        kernels = [kernels]
    # furnsh() needs path strings
    spiceypy.furnsh([str(kernel) for kernel in kernels])

    # Remove all existing SPICE frame classes
    global _frame_registry
    if _frame_registry:
        log.info(f"Removing {len(_frame_registry)} existing SPICE frame classes")
        for spice_frame_name in _frame_registry.keys():
            frame_cls, center_cls = _frame_registry[spice_frame_name]
            _uninstall_frame_by_class(frame_cls, center_cls)
        _frame_registry.clear()

    # Remove all non-default SPICE center classes
    global _center_registry
    if len(_center_registry) > 1:
        log.info(f"Removing {len(_center_registry) - 1} existing SPICE origin classes")
        for spice_center_name, center_cls in _center_registry.items():
            if center_cls != ICRS:
                _uninstall_frame_by_class(center_cls, ICRS)
        _center_registry = {'SOLAR SYSTEM BARYCENTER': ICRS}

    # Generate all SPICE frame classes
    for class_num in _CLASS_TYPES.keys():
        frames = spiceypy.kplfrm(class_num)
        for frame_id in frames:
            install_frame(frame_id)


def install_frame(spice_frame):
    """
    Install a specified SPICE frame.

    Installing a SPICE frame creates a corresponding frame class. All frames
    defined in the kernel pool are already automatically installed in the call
    to :func:`~sunpy.coordinates.spice.initialize`, so this function is used to
    manually install built-in frames, namely inertial or body-fixed (PCK) frames.
    Some common built-in frames include 'IAU_SUN', 'IAU_EARTH', and 'ITRF93'.

    Parameters
    ----------
    spice_frame : `str`, `int`
        The SPICE frame name or frame ID to be installed.

    Examples
    --------
    .. minigallery:: sunpy.coordinates.spice.install_frame
    """
    if isinstance(spice_frame, str):
        frame_name = spice_frame.upper()
        frame_id = spiceypy.namfrm(frame_name)
        if frame_id == 0:
            raise ValueError(f"{frame_name} is not a valid SPICE frame name.")
    else:
        frame_id = spice_frame
        frame_name = spiceypy.frmnam(frame_id)
        if frame_name == '':
            raise ValueError(f"{frame_id} is not a valid SPICE frame ID.")

    if frame_name in _frame_registry:
        log.info(f"{frame_name} (frame_id) has already been installed.")
    else:
        _install_frame_by_id(frame_id)


@add_common_docstring(**_variables_for_parse_time_docstring())
def get_body(body, time, *, spice_frame='J2000', observer=None):
    """
    Get the location of a body via SPICE.

    Parameters
    ----------
    body : `int`, `str`
        The NAIF body ID, or a string that is resolvable to a body ID
    time : {parse_time_types}
        Time to use in a parse_time-compatible format.
    spice_frame : `str`
        The SPICE frame name to use for the returned coordinate. Defaults to
        ``'J2000'``, which is equivalent to Astropy's `~astropy.coordinates.ICRS`.
    observer : `~astropy.coordinates.SkyCoord`
        If `None`, the returned coordinate is the instantaneous or “true” location.
        If not `None`, the returned coordinate is the astrometric location (i.e.,
        accounts for light travel time to the specified observer).

    Examples
    --------
    .. minigallery:: sunpy.coordinates.spice.get_body
    """
    body_name = spiceypy.bodc2n(body) if isinstance(body, int) else body
    obstime = parse_time(time)
    et = _convert_to_et(obstime)

    frame_center = spiceypy.frinfo(spiceypy.namfrm(spice_frame))[0]

    if observer is None:
        pos = spiceypy.spkpos(body_name,
                              et,
                              spice_frame,
                              'NONE',
                              spiceypy.bodc2n(frame_center))[0] << u.km
    else:
        obspos = observer.icrs.cartesian.xyz.to_value('km')
        pos, lt = spiceypy.spkcpo(body_name,
                                  et,
                                  spice_frame,
                                  'OBSERVER',
                                  'CN',
                                  obspos,
                                  'SSB',
                                  'J2000')
        log.info(f"Apparent body location accounts for {lt:.2f} seconds of light travel time")

        pos = pos[:3] << u.km
        if spice_frame != 'J2000':
            shift = spiceypy.spkpos(spiceypy.bodc2n(frame_center),
                                    et,
                                    'J2000',
                                    'NONE',
                                    'SSB')[0]
            obspos -= shift

            matrix = spiceypy.pxform('J2000', spice_frame, _convert_to_et(obstime))
            obspos = matrix @ obspos
        pos += obspos << u.km

    frame_name = 'icrs' if spice_frame == 'J2000' else _astropy_frame_name(spice_frame)

    return SkyCoord(CartesianRepresentation(pos.T), frame=frame_name, obstime=obstime)


@add_common_docstring(**_variables_for_parse_time_docstring())
def get_fov(instrument, time, *, resolution=100):
    """
    Get the field of view (FOV) for an instrument via SPICE.

    Rectangular and polygonal FOVs are represented by their vertices. Circular FOVs
    are approximated by a series of points. This function does not yet support
    elliptical FOVs.

    .. note::
        The FOV determined from SPICE kernels may not be as accurate as the FOV
        obtained from other sources of information, particularly if the instrument
        is an imager.

    Parameters
    ----------
    instrument : `int`, `str`
        The NAIF ID for the instrument, or a string that is resolvable to an
        instrument ID
    time : {parse_time_types}
        Time to use in a parse_time-compatible format.
    resolution : `int`
        Number of points to use for a circular FOV. Defaults to 100.

    Examples
    --------
    .. minigallery:: sunpy.coordinates.spice.get_fov
    """
    instrument_name = spiceypy.bodc2n(instrument) if isinstance(instrument, int) else instrument
    obstime = parse_time(time)

    fov_shape, spice_frame, boresight, num_vectors, vectors = spiceypy.getfvn(instrument_name, 1000)

    if fov_shape == "ELLIPSE":
        raise ValueError("Elliptical FOVs are not yet supported.")

    # If circular FOV, we have only one vector, so we have to rotate for the rest of the points
    if fov_shape == "CIRCLE":
        angles = np.arange(0, 1, 1/resolution) * 360*u.deg
        matrix = rotation_matrix(angles, axis=boresight)
        vectors = matrix @ vectors[0]
        num_vectors = resolution

    # The FOV vectors need to be replicated if obstime is a time array
    if not obstime.isscalar:
        vectors = np.broadcast_to(vectors[:, np.newaxis, :], (num_vectors, *obstime.shape, 3))
        obstime = Time(np.tile(obstime, num_vectors)).reshape((num_vectors, *obstime.shape)).T

    fov = SkyCoord(CartesianRepresentation(vectors.T),
                   frame=_frame_registry[spice_frame][0],
                   obstime=obstime,
                   representation_type='unitspherical')
    return fov


@add_common_docstring(**_variables_for_parse_time_docstring())
def get_rotation_matrix(source_frame, target_frame, from_time, to_time=None):
    """
    Get the rotation matrix between the orientations of two SPICE frames.

    Parameters
    ----------
    source_frame : `str`
        The source frame specified by SPICE frame name.

    target_frame : `str`
        The target frame specified by SPICE frame name.

    from_time : {parse_time_types}
        The time of the source frame.

    to_time : {parse_time_types}
        The time of the target frame. Defaults to ``None``, which means
        ``from_time`` is used.

    Returns
    -------
    `~numpy.ndarray`
        A 3x3 rotation matrix for the change in orientation.

    Examples
    --------
    >>> from sunpy.coordinates.spice import get_rotation_matrix
    >>> source_frame = "J2000"
    >>> target_frame = "Galactic"
    >>> from_time = '2001-01-01T00:00:00'
    >>> rotation_matrix = get_rotation_matrix(source_frame, target_frame, from_time)
    >>> rotation_matrix
    array([[-0.05487554, -0.8734371 , -0.48383499],
           [ 0.49410945, -0.44482959,  0.74698225],
           [-0.86766614, -0.19807639,  0.45598379]])

    This rotation matrix can be used to re-orient a vector (field), e.g.:

    >>> vec_components = [1, 0, 0] * u.T
    >>> transformed_matrix = rotation_matrix @ vec_components
    >>> transformed_matrix
    <Quantity [-0.05487554, 0.49410945, -0.86766614] T>
    """
    source_frame_spice = source_frame.upper()
    target_frame_spice = target_frame.upper()

    from_time = parse_time(from_time)
    to_time = parse_time(to_time) if to_time else from_time

    from_time_et = _convert_to_et(from_time)
    to_time_et = _convert_to_et(to_time)

    # First transformation: from source frame at from_time to J2000
    from_source_to_j2000 = spiceypy.sxform(source_frame_spice, "J2000", from_time_et)

    # Second transformation: from J2000 at to_time to target frame
    from_j2000_to_target = spiceypy.sxform("J2000", target_frame_spice, to_time_et)

    # Combine: source -> J2000 -> target
    combined_transform = spiceypy.mxmg(from_j2000_to_target, from_source_to_j2000)

    # Extract the rotation matrix (upper left 3x3 block)
    rotation_matrix = combined_transform[:3, :3]

    return rotation_matrix



import pytest

import astropy.units as u

from sunpy.net import attr
from sunpy.net import attrs as a
from sunpy.time import parse_time


def test_time_subclass():
    class NewTime(a.Time):
        pass

    assert isinstance(NewTime("2020/01/01", "2020/01/02") & a.Time("2020/02/02", "2020/02/03"), attr.AttrAnd)


def test_attrs_time():
    times = a.Time("2020/10/01T00:00", "2020/10/01T00:00")
    assert times.start == parse_time("2020/10/01T00:00")
    assert times.end == parse_time("2020/10/01T00:00")


def test_wavelength_attr():
    # Wavelength converts the input to be either kHz, keV or Angstrom
    wave = a.Wavelength(1*u.Mm, 2*u.Mm)
    assert u.allclose(wave.min, 1.e+16*u.Angstrom)
    assert wave.min.unit == u.Angstrom
    assert u.allclose(wave.max, 2.e+16*u.Angstrom)
    assert wave.max.unit == u.Angstrom
    assert wave.unconverted_value == (1*u.Mm, 2*u.Mm)


def test_instrument_show_in_notebook(mocker):
    pytest.importorskip("itables")
    mock_datagrid =  mocker.patch("itables.show")
    a.Instrument.show_in_notebook()
    mock_datagrid.assert_called_once()



"""
=======================================
Saving and loading sunpy Maps with FITS
=======================================

In this example we are going to look at how to save and load a
`~sunpy.map.GenericMap` as FITS files.

FITS is a modern file format designed to meet the needs of the astronomy
community. It has deep integration with Python, SunPy and Astropy as well as
IDL and many other languages.

Here, even though we will be working with `~sunpy.map.sources.sdo.AIAMap`
specifically, the process can be extended to any `~sunpy.map.GenericMap`.
"""
import astropy.units as u

import sunpy.data.sample
import sunpy.map

################################################################################
# We begin by creating an `~sunpy.map.sources.sdo.AIAMap` object using the
# sample data.

aia_map = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)
aia_map.peek(clip_interval=(1, 99.99)*u.percent)

################################################################################
# We can now save this object to a FITS file to use later. Saving it like this
# allows us to preserve all of the metadata of the object along with the actual
# array data. When we load the FITS file again, we get an identical
# `~sunpy.map.sources.sdo.AIAMap` object.
#
# All changes to `~sunpy.map.GenericMap` are saved within the FITS file. For example,
# if we rotate the image by 45 degrees and then save it to a FITS file, these changes
# are reflected in both the image data and metadata when we load the data back in.

aia_map = aia_map.rotate(45*u.deg)
# Please be aware that if you try to save this twice,
# it will throw an exception rather than overwriting the file.
aia_map.save('aia_map.fits')

################################################################################
# This FITS file is portable and can be safely loaded by any FITS reader.
# Using `~sunpy.map.Map`, we can easily read our rotated image back in:

aia_map_from_fits = sunpy.map.Map('aia_map.fits')
aia_map_from_fits.peek(clip_interval=(1, 99.99)*u.percent)



"""
Test cases for SDO HMIMap subclass.
"""
import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map import Map
from sunpy.map.sources.sdo import HMIMap
from .helpers import _test_private_date_setters


@pytest.fixture
def hmi_map():
    return Map(get_test_filepath('resampled_hmi.fits'))


@pytest.fixture
def hmi_bharp_map():
    return get_dummy_map_from_header(get_test_filepath('hmi_bharp_vlos_mag.header'))


@pytest.fixture
def hmi_cea_sharp_map():
    return get_dummy_map_from_header(get_test_filepath('hmi_cea_sharp_magnetogram.header'))


@pytest.fixture
def hmi_sharp_map():
    return get_dummy_map_from_header(get_test_filepath('hmi_sharp_magnetogram.header'))


def test_fitstoHMI(hmi_map, hmi_bharp_map, hmi_cea_sharp_map, hmi_sharp_map):
    """Tests the creation of HMIMap using FITS."""
    assert isinstance(hmi_map, HMIMap)
    assert isinstance(hmi_bharp_map, HMIMap)
    assert isinstance(hmi_cea_sharp_map, HMIMap)
    assert isinstance(hmi_sharp_map, HMIMap)


def test_is_datasource_for(hmi_map, hmi_bharp_map, hmi_cea_sharp_map, hmi_sharp_map):
    """Test the is_datasource_for method of HMIMap.
    Note that header data to be provided as an argument
    can be a MetaDict object, which in this case is
    hmi.meta."""
    assert hmi_map.is_datasource_for(hmi_map.data, hmi_map.meta)
    assert hmi_bharp_map.is_datasource_for(hmi_bharp_map.data, hmi_bharp_map.meta)
    assert hmi_cea_sharp_map.is_datasource_for(hmi_cea_sharp_map.data, hmi_cea_sharp_map.meta)
    assert hmi_sharp_map.is_datasource_for(hmi_sharp_map.data, hmi_sharp_map.meta)


def test_reference_date(hmi_map, hmi_bharp_map, hmi_cea_sharp_map, hmi_sharp_map):
    assert hmi_map.reference_date.isot == "2014-03-01T00:01:25.000"
    assert hmi_bharp_map.reference_date.isot == "2014-06-09T23:48:07.532"
    assert hmi_cea_sharp_map.reference_date.isot == "2024-06-28T00:00:08.212"
    assert hmi_sharp_map.reference_date.isot == "2024-06-28T00:00:08.212"


def test_date(hmi_map, hmi_bharp_map, hmi_cea_sharp_map, hmi_sharp_map):
    assert hmi_map.date.isot == "2014-03-01T00:00:27.900"
    assert hmi_bharp_map.date.isot == "2014-06-09T23:46:25.000"
    assert hmi_cea_sharp_map.date.isot == "2024-06-27T23:58:46.200"
    assert hmi_sharp_map.date.isot == "2024-06-27T23:58:46.200"


def test_private_date_setters(hmi_map, hmi_bharp_map, hmi_cea_sharp_map, hmi_sharp_map):
    _test_private_date_setters(hmi_map)
    _test_private_date_setters(hmi_bharp_map)
    _test_private_date_setters(hmi_cea_sharp_map)
    _test_private_date_setters(hmi_sharp_map)


def test_observatory(hmi_map, hmi_bharp_map, hmi_cea_sharp_map, hmi_sharp_map):
    """Tests the observatory property of the HMIMap object."""
    assert hmi_map.observatory == "SDO"
    assert hmi_bharp_map.observatory == "SDO"
    assert hmi_cea_sharp_map.observatory == "SDO"
    assert hmi_sharp_map.observatory == "SDO"


def test_measurement(hmi_map, hmi_bharp_map, hmi_cea_sharp_map, hmi_sharp_map):
    """Tests the measurement property of the HMIMap object."""
    assert hmi_map.measurement == "continuum"
    assert hmi_bharp_map.measurement == "observable"
    assert hmi_cea_sharp_map.measurement == "observable"
    assert hmi_sharp_map.measurement == "observable"


def test_wavelength(hmi_map, hmi_bharp_map, hmi_cea_sharp_map, hmi_sharp_map):
    assert hmi_map.wavelength == 6173  # The test file actually has an "" WAVEUNIT key
    assert hmi_bharp_map.wavelength == 6173 * u.AA
    assert hmi_cea_sharp_map.wavelength == 6173 * u.AA
    assert hmi_sharp_map.wavelength == 6173 * u.AA


def test_unit(hmi_cea_sharp_map, hmi_sharp_map):
    """Test that HMI source handles units not in FITS standard"""
    assert hmi_cea_sharp_map.unit == u.Unit('Mx/cm2')
    assert hmi_sharp_map.unit == u.Unit('Mx/cm2')


def test_wcs(hmi_map, hmi_bharp_map, hmi_cea_sharp_map, hmi_sharp_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    hmi_map.pixel_to_world(0*u.pix, 0*u.pix)
    hmi_bharp_map.pixel_to_world(0*u.pix, 0*u.pix)
    hmi_cea_sharp_map.pixel_to_world(0*u.pix, 0*u.pix)
    hmi_sharp_map.pixel_to_world(0*u.pix, 0*u.pix)




import warnings
from contextlib import nullcontext

import numpy as np
import pytest

import astropy.units as u
from astropy.coordinates import (
    CartesianRepresentation,
    SkyCoord,
    SphericalRepresentation,
    UnitSphericalRepresentation,
)
from astropy.tests.helper import assert_quantity_allclose

from sunpy import sun
from sunpy.coordinates import PlanarScreen, SphericalScreen, propagate_with_solar_surface
from sunpy.coordinates.frames import (
    Geomagnetic,
    Heliocentric,
    HeliographicCarrington,
    HeliographicStonyhurst,
    Helioprojective,
    HelioprojectiveRadial,
)
from sunpy.coordinates.sun import angular_radius
from sunpy.time import parse_time
from sunpy.util.exceptions import SunpyUserWarning

RSUN_METERS = sun.constants.get('radius').si.to(u.m)
DSUN_METERS = sun.constants.get('mean distance').si.to(u.m)


def init_frame(frame, args, kwargs):
    if args and kwargs:
        return frame(*args, **kwargs)
    elif args:
        return frame(*args)
    elif kwargs:
        return frame(**kwargs)


"""
These are common 2D params, kwargs are frame specific
"""
two_D_parameters = [
    ([0 * u.deg, 0 * u.arcsec], None),
    ([0 * u.deg, 0 * u.arcsec], {'obstime': '2011/01/01T00:00:00'}),
    ([UnitSphericalRepresentation(0 * u.deg, 0 * u.arcsec)], None),
    ([UnitSphericalRepresentation(0 * u.deg, 0 * u.arcsec)], {'obstime': '2011/01/01T00:00:00'}),
    ([0 * u.deg, 0 * u.arcsec], {'representation_type': 'unitspherical'})
]
"""
These are common 3D params, kwargs are frame specific
"""
three_D_parameters = [
    ([0 * u.deg, 0 * u.arcsec, 1 * u.Mm], None),
    ([0 * u.deg, 0 * u.arcsec, 1 * u.Mm], {'obstime': '2011/01/01T00:00:00'}),
    ([0 * u.deg, 0 * u.arcsec, 1 * u.Mm], {'representation_type': 'spherical'}),
    ([SphericalRepresentation(0 * u.deg, 0 * u.arcsec, 1 * u.Mm)],
     None),
    ([SphericalRepresentation(0 * u.deg, 0 * u.arcsec, 1 * u.Mm)], None), (
        [SphericalRepresentation(0 * u.deg, 0 * u.arcsec, 1 * u.Mm)],
        {'obstime': '2011/01/01T00:00:00'})
]

# ==============================================================================
# Helioprojective Tests
# ==============================================================================


@pytest.mark.parametrize(('args', 'kwargs'),
                         two_D_parameters + [(None, {'Tx': 0 * u.deg,
                                                     'Ty': 0 * u.arcsec})])
def test_create_hpc_2d(args, kwargs):
    hpc1 = init_frame(Helioprojective, args, kwargs)

    # Check we have the right class!
    assert isinstance(hpc1, Helioprojective)
    rep_kwarg = kwargs.get('representation_type', None) if kwargs else None

    if rep_kwarg and rep_kwarg == 'unitspherical':
        # Check that we have a unitspherical representation
        assert isinstance(hpc1._data, UnitSphericalRepresentation)
    else:
        # Check that we have a 2D wrap180 representation
        assert isinstance(hpc1._data, UnitSphericalRepresentation)

    # Check the attrs are correct
    assert hpc1.Tx == 0 * u.arcsec
    assert hpc1.Ty == 0 * u.arcsec

    # Check the attrs are in the correct default units
    assert hpc1.Tx.unit is u.arcsec
    assert hpc1.Ty.unit is u.arcsec


@pytest.mark.parametrize(
    ('args', 'kwargs'),
    three_D_parameters + [(None, {'Tx': 0 * u.deg,
                                  'Ty': 0 * u.arcsec,
                                  'distance': 1 * u.Mm}),
                          ([0 * u.deg, 0 * u.arcsec], {'distance': 1 * u.Mm})])
def test_create_3d(args, kwargs):
    hpc1 = init_frame(Helioprojective, args, kwargs)

    # Check we have the right class!
    assert isinstance(hpc1, Helioprojective)
    rep_kwarg = kwargs.get('representation_type', None) if kwargs else None

    if rep_kwarg and rep_kwarg == 'spherical':
        # Check that we have a unitspherical representation
        assert isinstance(hpc1._data, SphericalRepresentation)
    else:
        # Check that we have a 2D wrap180 representation
        assert isinstance(hpc1._data, SphericalRepresentation)

    # Check the attrs are correct
    assert hpc1.Tx == 0 * u.arcsec
    assert hpc1.Ty == 0 * u.arcsec
    assert hpc1.distance == 1 * u.Mm

    # Check the attrs are in the correct default units
    assert hpc1.Tx.unit is u.arcsec
    assert hpc1.Ty.unit is u.arcsec
    assert hpc1.distance.unit is u.Mm


def test_cart_init():
    hpc1 = Helioprojective(CartesianRepresentation(0 * u.km, 0 * u.km, 1 *
                                                   u.Mm))

    assert isinstance(hpc1, Helioprojective)
    assert isinstance(hpc1._data, CartesianRepresentation)


# Test HPC Calculate Distance
def test_hpc_distance():
    hpc1 = Helioprojective(0 * u.deg, 0 * u.arcsec,
                           observer=HeliographicStonyhurst(0*u.deg, 0*u.deg, 1*u.AU))

    assert isinstance(hpc1, Helioprojective)
    # Check that we have a 2D wrap180 representation
    assert isinstance(hpc1._data, UnitSphericalRepresentation)

    # Check the attrs are correct
    assert hpc1.Tx == 0 * u.arcsec
    assert hpc1.Ty == 0 * u.arcsec

    hpc2 = hpc1.make_3d()

    assert isinstance(hpc2._data, SphericalRepresentation)

    # Check the attrs are correct
    assert hpc2.Tx == 0 * u.arcsec
    assert hpc2.Ty == 0 * u.arcsec
    assert_quantity_allclose(hpc2.distance, DSUN_METERS - RSUN_METERS)


def test_hpc_distance_cartesian():
    # Test detection of distance in other representations
    hpc1 = Helioprojective(CartesianRepresentation(0 * u.km, 0 * u.km, 1 * u.Mm))

    assert isinstance(hpc1, Helioprojective)
    assert isinstance(hpc1._data, CartesianRepresentation)

    assert hpc1.make_3d() is hpc1


def test_hpc_distance_off_limb():
    hpc1 = Helioprojective(1500 * u.arcsec, 0 * u.arcsec,
                           observer=HeliographicStonyhurst(0*u.deg, 0*u.deg, 1*u.AU))

    assert isinstance(hpc1, Helioprojective)
    # Check that we have a 2D wrap180 representation
    assert isinstance(hpc1._data, UnitSphericalRepresentation)

    # Check the attrs are correct
    assert hpc1.Tx == 1500 * u.arcsec
    assert hpc1.Ty == 0 * u.arcsec

    with pytest.warns(SunpyUserWarning, match="is all NaNs"):
        hpc2 = hpc1.make_3d()
    assert isinstance(hpc2._data, SphericalRepresentation)
    # Check the attrs are correct
    assert hpc2.Tx == 1500 * u.arcsec
    assert hpc2.Ty == 0 * u.arcsec
    assert_quantity_allclose(hpc2.distance, u.Quantity(np.nan, u.km))


def test_hpc_distance_3D():
    hpc1 = Helioprojective(1500 * u.arcsec, 0 * u.arcsec, 100 * u.Mm)

    assert isinstance(hpc1, Helioprojective)
    # Check that we have a 2D wrap180 representation
    assert isinstance(hpc1._data, SphericalRepresentation)

    # Check the attrs are correct
    assert hpc1.Tx == 1500 * u.arcsec
    assert hpc1.Ty == 0 * u.arcsec

    hpc2 = hpc1.make_3d()

    assert hpc2 is hpc1


def test_wrapping_on():
    hpc1 = Helioprojective(359.9*u.deg, 10*u.deg)
    assert_quantity_allclose(hpc1.Tx, -0.1*u.deg)
    assert_quantity_allclose(hpc1.Tx.wrap_angle, 180*u.deg)


def test_wrapping_off():
    hpc1 = Helioprojective(359.9*u.deg, 10*u.deg, wrap_longitude=False)
    assert_quantity_allclose(hpc1.Tx, 359.9*u.deg)
    assert_quantity_allclose(hpc1.Tx.wrap_angle, 360*u.deg)


def test_hpc_default_observer():
    # Observer is considered default if it hasn't been specified *and* if obstime isn't specified
    hpc = Helioprojective(0*u.arcsec, 0*u.arcsec)
    assert hpc.is_frame_attr_default('observer')

    hpc = Helioprojective(0*u.arcsec, 0*u.arcsec, obstime='2019-06-01')
    assert not hpc.is_frame_attr_default('observer')


def test_hpc_low_precision_float_warning():
    with np.errstate(over='ignore'):
        hpc = Helioprojective(u.Quantity(0, u.deg, dtype=np.float32),
                              u.Quantity(0, u.arcsec, dtype=np.float16),
                              observer=HeliographicStonyhurst(0*u.deg, 0*u.deg, 1*u.AU))

        with pytest.warns(SunpyUserWarning, match="Tx is float32, and Ty is float16"):
            hpc.make_3d()


def test_hpc_obstime_from_observer():
    # Test that observer.obstime is used for obstime if obstime is not provided
    observer = HeliographicStonyhurst(obstime='2023-09-08')
    hpc = Helioprojective(observer=observer)
    assert hpc.obstime == observer.obstime

    # Test that obstime is None if observer does not have an obstime
    hpc = Helioprojective(observer='earth')
    assert hpc.obstime is None


def test_hpc_is_visible_2d():
    hpc = Helioprojective(2000*u.arcsec, 2000*u.arcsec,
                          observer='earth', obstime='2023-08-03')
    assert hpc.is_visible()


def test_hpc_is_visible():
    hpc = Helioprojective([0]*2*u.arcsec, [0]*2*u.arcsec, [0.5, 1.5]*u.AU,
                          observer='earth', obstime='2023-08-03')
    assert (hpc.is_visible() == [True, False]).all()


def test_hpc_is_visible_tolerance():
    hpc = Helioprojective(200*u.arcsec, 0*u.arcsec,
                          observer='earth', obstime='2023-08-03').make_3d()

    # Due to the limitations of numerical precision, the coordinate may be computed to be slightly
    # below the solar surface, and thus may be invisible when the tolerance is set to zero
    if hpc.is_visible(tolerance=0*u.m):
        pytest.skip("Test already passes prior to increasing the tolerance.")

    assert hpc.is_visible(tolerance=1*u.m)


# ==============================================================================
# Helioprojective Radial Tests
# ==============================================================================

@pytest.mark.parametrize(('args', 'kwargs'),
                         two_D_parameters + [(None, {'psi': 0 * u.deg,
                                                     'delta': 0 * u.arcsec})])
def test_create_hpr_2d(args, kwargs):
    hpr1 = init_frame(HelioprojectiveRadial, args, kwargs)

    assert isinstance(hpr1, HelioprojectiveRadial)
    assert isinstance(hpr1._data, UnitSphericalRepresentation)

    assert hpr1.psi.unit is u.deg
    assert hpr1.delta.unit is u.deg
    assert_quantity_allclose(hpr1.psi, 0*u.deg)
    assert_quantity_allclose(hpr1.delta, 0*u.deg)

    assert_quantity_allclose(hpr1.theta, 90*u.deg)


@pytest.mark.parametrize(
    ('args', 'kwargs'),
    three_D_parameters + [(None, {'psi': 0 * u.deg,
                                  'delta': 0 * u.arcsec,
                                  'distance': 1 * u.Mm}),
                          ([0 * u.deg, 0 * u.arcsec], {'distance': 1 * u.Mm})])
def test_create_hpr_3d(args, kwargs):
    hpr1 = init_frame(HelioprojectiveRadial, args, kwargs)

    assert isinstance(hpr1, HelioprojectiveRadial)
    assert isinstance(hpr1._data, SphericalRepresentation)

    assert hpr1.psi.unit is u.deg
    assert hpr1.delta.unit is u.deg
    assert hpr1.distance.unit is u.Mm
    assert_quantity_allclose(hpr1.psi, 0*u.deg)
    assert_quantity_allclose(hpr1.delta, 0*u.deg)
    assert_quantity_allclose(hpr1.distance, 1*u.Mm)

    assert_quantity_allclose(hpr1.theta, 90*u.deg)

    # Since hpr1 is already 3D, make_3d() should simply return the original object
    hpr2 = hpr1.make_3d()
    assert hpr2 is hpr1


def test_hpr_distance():
    hpr1 = HelioprojectiveRadial(0*u.deg, -90*u.deg,
                                 observer=HeliographicStonyhurst(0*u.deg, 0*u.deg, 1*u.AU))

    hpr2 = hpr1.make_3d()

    assert_quantity_allclose(hpr2.psi, 0*u.deg)
    assert_quantity_allclose(hpr2.delta, -90*u.deg)
    assert_quantity_allclose(hpr2.distance, DSUN_METERS - RSUN_METERS)


# ==============================================================================
# ## Heliographic Tests
# ==============================================================================

def test_HEEQ_creation():
    # Smoke test to make sure HEEQ constructors work fine
    _ = HeliographicStonyhurst(lon=0*u.deg, lat=90*u.deg,
                               obstime=parse_time('2018-12-21'))
    _ = HeliographicStonyhurst(lon=0*u.deg, lat=90*u.deg, radius=1*u.km,
                               obstime=parse_time('2018-12-21'))
    _ = HeliographicStonyhurst(x=1*u.km, y=1*u.km, z=1*u.km,
                               obstime=parse_time('2018-12-21'),
                               representation_type='cartesian')


@pytest.mark.parametrize('frame',
                         [HeliographicStonyhurst, HeliographicCarrington])
@pytest.mark.parametrize(('args', 'kwargs'), two_D_parameters +
                         [(None, {'lat': 0*u.deg, 'lon': 0*u.arcsec})])
def test_create_hgs_2d(frame, args, kwargs):
    hgs1 = init_frame(frame, args, kwargs)

    # Check we have the right class!
    assert isinstance(hgs1, frame)
    # Check that we have a 2D representation
    assert isinstance(hgs1._data, UnitSphericalRepresentation)

    # Check the attrs are correct
    assert hgs1.lon == 0 * u.deg
    assert hgs1.lat == 0 * u.deg

    # Check the attrs are in the correct default units
    assert hgs1.lon.unit is u.deg
    assert hgs1.lat.unit is u.deg

    # Test the value of the rsun frame attribute
    assert_quantity_allclose(hgs1.rsun, sun.constants.radius)

    # Test conversion to 3D
    hgs_3d = hgs1.make_3d()
    assert_quantity_allclose(hgs_3d.lon, hgs1.lon)
    assert_quantity_allclose(hgs_3d.lat, hgs1.lat)
    assert_quantity_allclose(hgs_3d.radius, hgs1.rsun)


@pytest.mark.parametrize('frame',
                         [HeliographicStonyhurst, HeliographicCarrington])
@pytest.mark.parametrize(
    ('args', 'kwargs'),
    three_D_parameters + [(None, {'lat': 0 * u.deg,
                                  'lon': 0 * u.arcsec,
                                  'radius': 1 * u.Mm}),
                          ([0 * u.deg, 0 * u.arcsec], {'radius': 1 * u.Mm})])
def test_create_hgs_3d(frame, args, kwargs):
    hgs1 = init_frame(frame, args, kwargs)

    # Check we have the right class!
    assert isinstance(hgs1, frame)

    rep_kwarg = kwargs.get('representation_type', None) if kwargs else None

    if rep_kwarg == 'spherical':
        assert isinstance(hgs1._data, SphericalRepresentation)
    else:
        # Check Carrington first because it's a subclass of Stonyhurst
        if isinstance(hgs1, HeliographicCarrington):
            # Check that we have a 2D wrap180 representation
            assert isinstance(hgs1._data, SphericalRepresentation)
        elif isinstance(hgs1, HeliographicStonyhurst):
            # Check that we have a 2D wrap180 representation
            assert isinstance(hgs1._data, SphericalRepresentation)

    # Check the attrs are correct
    assert hgs1.lon == 0 * u.deg
    assert hgs1.lat == 0 * u.deg
    assert hgs1.radius == 1 * u.Mm

    # Check the attrs are in the correct default units
    assert hgs1.lon.unit is u.deg
    assert hgs1.lat.unit is u.deg
    assert hgs1.radius.unit is u.Mm


def test_hgs_cart_init():
    hpc1 = HeliographicStonyhurst(CartesianRepresentation(0 * u.km,
                                                          0 * u.km,
                                                          1 * u.Mm))

    assert isinstance(hpc1, HeliographicStonyhurst)
    assert isinstance(hpc1._data, CartesianRepresentation)


def test_hgs_wrapping_on():
    hpc1 = HeliographicStonyhurst(350*u.deg, 10*u.deg)
    assert_quantity_allclose(hpc1.lon, -10*u.deg)
    assert_quantity_allclose(hpc1.lon.wrap_angle, 180*u.deg)


def test_hgs_wrapping_off():
    hpc1 = HeliographicStonyhurst(350*u.deg, 10*u.deg, wrap_longitude=False)
    assert_quantity_allclose(hpc1.lon, 350*u.deg)
    assert_quantity_allclose(hpc1.lon.wrap_angle, 360*u.deg)


def test_hgc_wrapping_360():
    hpc1 = HeliographicCarrington(350*u.deg, 10*u.deg)
    assert_quantity_allclose(hpc1.lon, 350*u.deg)
    assert_quantity_allclose(hpc1.lon.wrap_angle, 360*u.deg)


# ==============================================================================
# ## Heliocentric Tests
# ==============================================================================
@pytest.mark.parametrize(
    ('args', 'kwargs'),
    [((10 * u.km, 10 * u.km, 10 * u.km), None), (None, {'x': 10 * u.km,
                                                        'y': 10 * u.km,
                                                        'z': 10 * u.km}),
     ([CartesianRepresentation(10 * u.km, 10 * u.km, 10 * u.km)], None),
     ([CartesianRepresentation(10 * u.km, 10 * u.km, 10 * u.km)],
      {'obstime': '2011/01/01T00:00:00'})])
def test_create_hcc_3d(args, kwargs):
    hcc = init_frame(Heliocentric, args, kwargs)

    assert isinstance(hcc, Heliocentric)

    assert isinstance(hcc._data, CartesianRepresentation)

    assert hcc.x == 10 * u.km
    assert hcc.y == 10 * u.km
    assert hcc.z == 10 * u.km

    # Check the attrs are in the correct default units
    assert hcc.x.unit is u.km
    assert hcc.y.unit is u.km
    assert hcc.z.unit is u.km


def test_hcc_default_observer():
    # Observer is considered default if it hasn't been specified *and* if obstime isn't specified
    hcc = Heliocentric(0*u.AU, 0*u.AU, 0*u.AU)
    assert hcc.is_frame_attr_default('observer')

    hcc = Heliocentric(0*u.AU, 0*u.AU, 0*u.AU, obstime='2019-06-01')
    assert not hcc.is_frame_attr_default('observer')


@pytest.mark.parametrize(('x', 'y', 'psi'), [(0*u.km, -1*u.km, 270*u.deg),
                                             (0*u.km, 1*u.km, 90*u.deg),
                                             (-1*u.km, 0*u.km, 180*u.deg)])
def test_heliocentric_radial_psi(x, y, psi):
    # The cylindrical representation of HCC is Heliocentric Radial
    # Test that the `psi` component is represented as desired
    # The definition is shifted by 90 degrees relative to Thompson (2006)
    hcc = Heliocentric(CartesianRepresentation(x, y, 0*u.km), representation_type='cylindrical')

    assert_quantity_allclose(hcc.psi, psi)


# ==============================================================================
# Magnetic-model coordinate frame tests
# ==============================================================================


@pytest.mark.parametrize(('obstime', 'dipole_lonlat'),
                         [('2012-07-01', [-72.408328, 80.16423]*u.deg),
                          ('2012-08-01', [-72.415148, 80.169261]*u.deg),
                          ('2032-08-01', [-72.576963, 81.212062]*u.deg),
                          (['2012-07-01', '2012-08-01'], [[-72.408328, 80.16423], [-72.415148, 80.169261]]*u.deg),
                          (['2012-07-01', '2032-08-01'], [[-72.408328, 80.16423], [-72.576963, 81.212062]]*u.deg),
                          ('1899-01-01', ValueError),
                          (['1899-01-01', '2012-07-01'], ValueError)])
def test_magnetic_model_obstime(obstime, dipole_lonlat):
    frame = Geomagnetic(obstime=obstime, magnetic_model='igrf13')
    ctx = pytest.raises(ValueError, match="earlier than the year 1900") if dipole_lonlat is ValueError else nullcontext()
    with ctx:
        assert_quantity_allclose(frame.dipole_lonlat, dipole_lonlat.T)


def test_magnetic_model_default():
    # Also tests that no downloading happens because this test is not marked as remote
    obstime = '2012-07-01'
    frame_default = Geomagnetic(obstime=obstime)
    frame_igrf13 = Geomagnetic(obstime=obstime, magnetic_model='igrf13')

    assert_quantity_allclose(frame_igrf13.dipole_lonlat, [-72.408328, 80.16423]*u.deg)
    assert_quantity_allclose(frame_default.dipole_lonlat, frame_igrf13.dipole_lonlat)


@pytest.mark.remote_data
@pytest.mark.parametrize(('magnetic_model', 'obstime', 'dipole_lonlat'),
                         [('igrf12', '2012-07-01', [-72.414318, 80.16354]*u.deg),
                          ('igrf11', '2006-01-01', [-71.886023, 79.801523]*u.deg),
                          ('igrf10', '2006-01-01', [-71.822653, 79.785185]*u.deg)])
def test_magnetic_model_with(magnetic_model, obstime, dipole_lonlat):
    frame = Geomagnetic(magnetic_model=magnetic_model, obstime=obstime)
    assert_quantity_allclose(frame.dipole_lonlat, dipole_lonlat)


# ==============================================================================
# SkyCoord Tests
# ==============================================================================


two_D_parameters = [
    ([0 * u.deg, 0 * u.arcsec], {}),
    ([[0, 1, 2, 3], [5, 6, 7, 8]], {'unit': u.deg}),
    ([0 * u.deg, 0 * u.arcsec], {'representation_type': SphericalRepresentation}),
    ([UnitSphericalRepresentation(0 * u.deg, 0 * u.arcsec)], {}),
    ([UnitSphericalRepresentation(0 * u.deg, 0 * u.arcsec)], {}),
    ([SphericalRepresentation(0 * u.deg, 0 * u.arcsec, 1*u.one)], {}),
]


@pytest.mark.parametrize(('args', 'kwargs'),
                         two_D_parameters + [([0 * u.deg, 0 * u.arcsec],
                                              {'representation_type': 'unitspherical'})])
def test_skycoord_hpc(args, kwargs):
    """
    Test that when instantiating a HPC frame with SkyCoord that make_3d
    still works.
    """
    sc = SkyCoord(*args, **kwargs, frame="helioprojective",
                  observer='earth', obstime="2011-01-01T00:00:00")
    # Test the transform to HGS because it will force a `make_3d` call.
    # Only 1 of the 7 parameterized arguments actually emits the warning
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=(
            'The conversion of these 2D helioprojective coordinates to '
        ), category=SunpyUserWarning)
        hgs = sc.transform_to("heliographic_stonyhurst")

    assert isinstance(hgs.frame, HeliographicStonyhurst)


def test_hgc_incomplete_observer():
    with pytest.raises(ValueError, match=r'Full 3D coordinate \(including radius\) must be specified'):
        SkyCoord(0*u.deg, 0*u.deg, frame="heliographic_carrington",
                 observer='self', obstime="2011-01-01T00:00:00")


def test_angular_radius():
    coord = Helioprojective(0*u.deg, 0*u.deg, 5*u.km, obstime="2010/01/01T00:00:00", observer="earth")
    assert_quantity_allclose(coord.angular_radius, angular_radius(coord.obstime))


def test_angular_radius_no_observer():
    coord = Helioprojective(0*u.deg, 0*u.deg, 5*u.km, obstime="2010/01/01T00:00:00", observer=None)
    with pytest.raises(ValueError, match=r"The observer must be defined, not `None`"):
        coord.angular_radius


def test_angular_radius_no_obstime():
    coord = Helioprojective(0*u.deg, 0*u.deg, 5*u.km, obstime=None, observer="earth")
    with pytest.raises(ValueError, match=r"The observer must be fully defined by specifying `obstime`."):
        coord.angular_radius


@pytest.fixture
def off_limb_coord():
    frame = Helioprojective(observer='earth', obstime='2020-01-01')
    return SkyCoord(Tx=[-1000, 300, 1000]*u.arcsec, Ty=[-1000, 300, 1000]*u.arcsec, frame=frame)


@pytest.fixture
def non_earth_coord():
    return SkyCoord(70*u.deg, 20*u.deg, 1*u.AU, obstime='2020-01-01', frame=HeliographicStonyhurst)


@pytest.mark.parametrize('screen_class', [
    SphericalScreen,
    PlanarScreen,
])
def test_screen_classes(off_limb_coord, screen_class):
    # Smoke test for spherical screen
    with pytest.warns(SunpyUserWarning, match='The conversion of these 2D helioprojective coordinates to 3D is all NaNs'):
            olc_3d = off_limb_coord[0].make_3d()
    assert np.isnan(olc_3d.distance).all()
    sph_screen = screen_class(off_limb_coord[0].observer)
    with sph_screen:
        olc_3d = off_limb_coord[0].make_3d()
    assert not np.isnan(olc_3d.distance).all()


@pytest.mark.parametrize(('only_off_disk', 'distance_from_center', 'distance'), [
    (False, 0*u.m, [0.98331616, 0.98329512, 0.98331616]*u.AU),
    (True, 0*u.m, [0.98331616, 0.97910333, 0.98331616]*u.AU),
    (False, 1*u.Rsun, [0.97866558, 0.97864465, 0.97866558]*u.AU),
])
def test_planar_screen(off_limb_coord, only_off_disk, distance_from_center, distance):
    with PlanarScreen(off_limb_coord.observer, distance_from_center=distance_from_center, only_off_disk=only_off_disk):
        olc_3d = off_limb_coord.make_3d()
    assert u.quantity.allclose(olc_3d.distance, distance)


@pytest.mark.parametrize(('only_off_disk', 'distance'), [
    (False, [0.98329304, 0.98329304, 0.98329304]*u.AU),
    (True, [0.98329304, 0.97910333, 0.98329304]*u.AU),
])
def test_spherical_screen(off_limb_coord, only_off_disk, distance):
    with SphericalScreen(off_limb_coord.observer, only_off_disk=only_off_disk):
        olc_3d = off_limb_coord.make_3d()
    assert u.quantity.allclose(olc_3d.distance, distance)


@pytest.mark.parametrize(('only_off_disk', 'distance_from_center', 'distance'), [
    (False, 0*u.m, [0.96419505, 0.98918004, 1.00321100]*u.AU),
    (True, 0*u.m, [0.96419505, 0.97910333, 1.00321100]*u.AU),
    (False, 1*u.Rsun, [0.94916557, 0.97376111, 0.98757335]*u.AU),
])
def test_planar_screen_askew(off_limb_coord, only_off_disk, distance_from_center, distance, non_earth_coord):
    with PlanarScreen(non_earth_coord, distance_from_center=distance_from_center, only_off_disk=only_off_disk):
        olc_3d = off_limb_coord.make_3d()
    assert u.quantity.allclose(olc_3d.distance, distance)


@pytest.mark.parametrize(('only_off_disk', 'distance'), [
    (False, [0.96348934, 0.98911699, 1.00251206]*u.AU),
    (True, [0.96348934, 0.97910333, 1.00251206]*u.AU),
])
def test_spherical_screen_askew(off_limb_coord, only_off_disk, distance, non_earth_coord):
    with SphericalScreen(non_earth_coord, only_off_disk=only_off_disk):
        olc_3d = off_limb_coord.make_3d()
    assert u.quantity.allclose(olc_3d.distance, distance)


@pytest.mark.parametrize(('screen', 'only_off_disk', 'distance'), [
    (SphericalScreen, False, [0.98405002, 0.98306592, 0.98253594]*u.AU),
    (SphericalScreen, True, [0.98405002, 0.97910333, 0.98253594]*u.AU),
    (PlanarScreen, False, [0.98407381, 0.98306806, 0.98255967]*u.AU),
    (PlanarScreen, True, [0.98407381, 0.97910333, 0.98255967]*u.AU),
])
def test_screen_plus_diffrot(off_limb_coord, screen, only_off_disk, distance):
    new_observer = HeliographicStonyhurst(0*u.deg, 0*u.deg, 1*u.AU, obstime=off_limb_coord.obstime + 1*u.day)
    with propagate_with_solar_surface(), screen(new_observer, only_off_disk=only_off_disk):
        olc_3d = off_limb_coord.make_3d()
    assert_quantity_allclose(olc_3d.distance, distance)


@pytest.mark.parametrize('only_off_disk', [False, True])
@pytest.mark.parametrize('screen', [SphericalScreen, PlanarScreen])
def test_screen_plus_diffrot_array_obstime(screen, only_off_disk):
    array_obstime = parse_time('2025-05-30') + np.arange(5) * u.day
    observer = HeliographicStonyhurst(0*u.deg, 0*u.deg, 1*u.AU, obstime=array_obstime)
    coord_2d = SkyCoord([-1000, -750, -500, -250, 0] * u.arcsec,
                        [600]*5 * u.arcsec,
                        frame=Helioprojective(observer=observer, obstime=array_obstime))

    with propagate_with_solar_surface(), screen(observer[0], only_off_disk=only_off_disk):
        # Confirm that array obstime works
        coord_3d = coord_2d.make_3d()

        # Confirm that calculated distances match the calculation when done individually
        for point_2d, point_3d in zip(coord_2d, coord_3d):
            point_2d_3d = point_2d.make_3d()
            assert_quantity_allclose(point_2d_3d.separation_3d(point_3d), 0*u.m, atol=1*u.m)



from xml.dom.minidom import Document, parseString
from xml.parsers.expat import ExpatError

import pytest

from sunpy.util import xml


def test_xml_to_dict1():
    """
    should return dict of xml string.
    """
    source_xml = "<outer>\
          <inner1>one</inner1>\
          <inner2>two</inner2>\
        </outer>"

    xml_dict = xml.xml_to_dict(source_xml)
    expected_dict = {'outer': {'inner2': 'two', 'inner1': 'one'}}

    assert xml_dict == expected_dict


def test_xml_to_dict2():
    """
    should return dict of xml string and if a tag is duplicated it takes the
    last one.
    """
    source_xml = "<outer>\
                    <inner1>one-one</inner1>\
                    <inner1>one-two</inner1>\
                    <inner2>two-one</inner2>\
                    <inner2>two-two</inner2>\
                 </outer>"

    xml_dict = xml.xml_to_dict(source_xml)
    expected_dict = {'outer': {'inner2': 'two-two', 'inner1': 'one-two'}}

    assert xml_dict == expected_dict


def test_xml_to_dict3():
    """
    should return dict of xml string with empty value if there are no inner
    elements.
    """
    source_xml = "<outer/>"

    xml_dict = xml.xml_to_dict(source_xml)
    expected_dict = {'outer': ''}

    assert xml_dict == expected_dict


def test_xml_to_dict4():
    """
    should return dict of xml string with empty value if there are no inner
    elements.
    """
    source_xml = "<outer></outer>"

    xml_dict = xml.xml_to_dict(source_xml)
    expected_dict = {'outer': ''}

    assert xml_dict == expected_dict


def test_xml_to_dict5():
    """
    should return dict of xml string with 2 layer nesting.
    """
    source_xml = "<outer>\
                    <mid1>\
                        <inner1>one-one</inner1>\
                    </mid1>\
                    <mid2>\
                        <inner2>two-one</inner2>\
                    </mid2>\
                 </outer>"

    xml_dict = xml.xml_to_dict(source_xml)
    expected_dict = {'outer': {'mid2': {'inner2': 'two-one'}, 'mid1': {'inner1': 'one-one'}}}

    assert xml_dict == expected_dict


def test_xml_to_dict6():
    """
    should return dict of xml string with 2 layer nesting and if a tag is
    duplicated it takes the last one.
    """
    source_xml = "<outer>\
                    <mid>\
                        <inner1>one-one</inner1>\
                    </mid>\
                    <mid>\
                        <inner2>two-one</inner2>\
                    </mid>\
                 </outer>"

    xml_dict = xml.xml_to_dict(source_xml)
    expected_dict = {'outer': {'mid': {'inner2': 'two-one'}}}

    assert xml_dict == expected_dict


def test_xml_to_dict7():
    """
    should raise TypeError when passed None.
    """
    assert pytest.raises(TypeError, xml.xml_to_dict, None)


def test_xml_to_dict8():
    """
    should raise TypeError when passed non string.
    """
    assert pytest.raises(TypeError, xml.xml_to_dict, 9)


def test_xml_to_dict9():
    """
    should raise ExpatError when passed empty string.
    """
    assert pytest.raises(ExpatError, xml.xml_to_dict, "")


def test_xml_to_dict10():
    """
    should raise ExpatError when passed space.
    """
    assert pytest.raises(ExpatError, xml.xml_to_dict, " ")


def test_get_node_text1():
    """
    should raise NotTextNodeError if there is a non text node.
    """
    doc = Document()
    outer = doc.createElement("outer")
    doc.appendChild(outer)
    pytest.raises(xml.NotTextNodeError, xml.get_node_text, doc)


def test_get_node_text2():
    """
    should return empty string for a node with no child nodes.
    """
    assert xml.get_node_text(Document()) == ""


def test_get_node_text3():
    """
    should return node text.
    """
    node = parseString("<outer>one</outer>")
    text_node = node.childNodes[0]

    assert xml.get_node_text(text_node) == "one"


def test_get_node_text4():
    """
    should raise AttributeError when sent None.
    """
    assert pytest.raises(AttributeError, xml.get_node_text, None)


def test_get_node_text5():
    """
    should raise AttributeError when sent wrong type.
    """
    assert pytest.raises(AttributeError, xml.get_node_text, "wrong type")


def test_node_to_dict1():
    """
    should return dict of node.
    """

    doc = Document()

    outer = doc.createElement("outer")
    doc.appendChild(outer)

    inner1 = doc.createElement("inner1")
    inner2 = doc.createElement("inner2")
    outer.appendChild(inner1)
    outer.appendChild(inner2)

    inner1_text = doc.createTextNode("one")
    inner2_text = doc.createTextNode("two")
    inner1.appendChild(inner1_text)
    inner2.appendChild(inner2_text)

    expected_dict = {'outer': {'inner2': 'two', 'inner1': 'one'}}
    xml_dict = xml.node_to_dict(doc)

    assert xml_dict == expected_dict


def test_node_to_dict2():
    """
    should return dict of node double nested.
    """

    doc = Document()

    outer = doc.createElement("outer")
    doc.appendChild(outer)

    mid1 = doc.createElement("mid1")
    outer.appendChild(mid1)
    mid2 = doc.createElement("mid2")
    outer.appendChild(mid2)

    inner1 = doc.createElement("inner1")
    inner2 = doc.createElement("inner2")
    mid1.appendChild(inner1)
    mid2.appendChild(inner2)

    inner1_text = doc.createTextNode("one")
    inner2_text = doc.createTextNode("two")
    inner1.appendChild(inner1_text)
    inner2.appendChild(inner2_text)

    expected_dict = {'outer': {'mid2': {'inner2': 'two'}, 'mid1': {'inner1': 'one'}}}
    xml_dict = xml.node_to_dict(doc)

    assert xml_dict == expected_dict


def test_node_to_dict3():
    """
    should return empty dict when sent empty doc.
    """
    expected_dict = {}
    xml_dict = xml.node_to_dict(Document())

    assert xml_dict == expected_dict


def test_node_to_dict4():
    """
    should raise AttributeError when sent wrong type.
    """
    assert pytest.raises(AttributeError, xml.node_to_dict, 9)


def test_node_to_dict5():
    """
    should raise AttributeError when sent None.
    """
    assert pytest.raises(AttributeError, xml.node_to_dict, None)


def test_with_multiple_children_in_list():
    """
    Setting the 'multiple' attribute of parent node should put child nodes in a
    list.
    """
    def getChild(lst_of_children, key, value):
        for child in lst_of_children:
            if child[key] == value:
                return child

        raise ValueError(f"No children with key {key} set to {value} found.")

    source = '''<?xml version="1.0" encoding="UTF-8"?>
                <Config>
                    <Name>With multiple children</Name>
                    <Children multiple="true">
                      <Child>
                        <Name>First Child</Name>
                        <Value>Value 1</Value>
                      </Child>
                      <Child>
                        <Name>Second Child</Name>
                        <Value>Value 2</Value>
                      </Child>
                    </Children>
                </Config>'''

    expected = {'Config': {'Children': [{'Name': 'First Child', 'Value': 'Value 1'},
                                        {'Name': 'Second Child', 'Value': 'Value 2'}],
                           'Name': 'With multiple children'}}

    actual = xml.xml_to_dict(source)

    assert len(expected['Config']) == len(actual['Config'])
    assert expected['Config']['Name'] == actual['Config']['Name']

    assert len(actual['Config']['Children']) == 2

    # As the child dictionaries are in lists we cannot be certain what order
    # they are in. Test individually.

    expected_children = expected['Config']['Children']
    actual_children = actual['Config']['Children']

    expected_first_child = getChild(expected_children, key='Name', value='First Child')
    actual_first_child = getChild(actual_children, key='Name', value='First Child')

    assert expected_first_child == actual_first_child

    expected_second_child = getChild(expected_children, key='Name', value='Second Child')
    actual_second_child = getChild(actual_children, key='Name', value='Second Child')

    assert expected_second_child == actual_second_child



"""
This package contains Python packages that are bundled with ``sunpy`` but are
external to ``sunpy``, and hence are developed in a separate source tree.
"""
from .appdirs import *  # NOQA
from .distro import *  # NOQA
from .inflect import *  # NOQA
from .parse import *  # NOQA



"""
Coordinate frames that are defined relative to other frames
"""

import astropy.units as u
from astropy.coordinates.attributes import Attribute, QuantityAttribute
from astropy.coordinates.baseframe import frame_transform_graph
from astropy.coordinates.transformations import FunctionTransform

from sunpy.time import parse_time
from sunpy.time.time import _variables_for_parse_time_docstring
from sunpy.util.decorators import add_common_docstring
from ._transformations import _transformation_debug
from .frames import HeliographicStonyhurst, SunPyBaseCoordinateFrame
from .offset_frame import NorthOffsetFrame

__all__ = ['NorthOffsetFrame', 'RotatedSunFrame']


# The code for NorthOffsetFrame currently lives in `offset_frame.py`
# This changes its module to this file so that the docs think that the class is local
NorthOffsetFrame.__module__ = __name__


_rotatedsun_cache = {}


def _make_rotatedsun_cls(framecls):
    """
    Create a new class that is the rotated-Sun frame for a specific class of
    base frame. If such a class has already been created for this frame, the
    same class will be returned.

    This function is necessary because frame transformations depend
    on connection between specific frame *classes*. So each type of frame
    needs its own distinct rotated-Sun frame class. This function generates
    just that class, as well as ensuring that only one example of such a class
    actually gets created in any given Python session.
    """
    # This code reuses significant code from Astropy's implementation of SkyOffsetFrame
    # See licenses/ASTROPY.rst

    if framecls in _rotatedsun_cache:
        return _rotatedsun_cache[framecls]

    members = {'__doc__': f'{RotatedSunFrame.__doc__}\n{framecls.__doc__}'}

    # Copy over the defaults from the input frame class
    attrs_to_copy = ['_default_representation',
                     '_default_differential',
                     '_frame_specific_representation_info']
    for attr in attrs_to_copy:
        members[attr] = getattr(framecls, attr)

    _RotatedSunFramecls = type(f'RotatedSun{framecls.__name__}', (RotatedSunFrame,), members)

    @frame_transform_graph.transform(FunctionTransform, _RotatedSunFramecls, _RotatedSunFramecls)
    @_transformation_debug(f"{_RotatedSunFramecls.__name__}->{_RotatedSunFramecls.__name__}")
    def rotatedsun_to_rotatedsun(from_rotatedsun_coord, to_rotatedsun_frame):
        """Transform between two rotated-Sun frames."""
        # This transform goes through the parent frames on each side.
        # from_frame -> HGS -> to_frame
        int_frame = HeliographicStonyhurst(obstime=from_rotatedsun_coord.base.obstime)
        int_coord = from_rotatedsun_coord.transform_to(int_frame)
        return int_coord.transform_to(to_rotatedsun_frame)

    @frame_transform_graph.transform(FunctionTransform, HeliographicStonyhurst, _RotatedSunFramecls)
    @_transformation_debug(f"HGS->{_RotatedSunFramecls.__name__}")
    def reference_to_rotatedsun(hgs_coord, rotatedsun_frame):
        int_frame = HeliographicStonyhurst(obstime=rotatedsun_frame.base.obstime)
        int_coord = hgs_coord.make_3d().transform_to(int_frame)  # obstime change handled here

        # Rotate the coordinate in HGS
        int_coord = int_coord._apply_diffrot(-rotatedsun_frame.duration,
                                             rotatedsun_frame.rotation_model)

        # Transform from HGS
        new_coord = int_coord.transform_to(rotatedsun_frame.base)
        return rotatedsun_frame.realize_frame(new_coord.data)

    @frame_transform_graph.transform(FunctionTransform, _RotatedSunFramecls, HeliographicStonyhurst)
    @_transformation_debug(f"{_RotatedSunFramecls.__name__}->HGS")
    def rotatedsun_to_reference(rotatedsun_coord, hgs_frame):
        # Transform to HGS
        from_coord = rotatedsun_coord.base.realize_frame(rotatedsun_coord.data)
        if hasattr(from_coord, 'make_3d'):
            from_coord = from_coord.make_3d()
        int_frame = HeliographicStonyhurst(obstime=rotatedsun_coord.base.obstime)
        int_coord = from_coord.transform_to(int_frame)

        # Rotate the coordinate in HGS
        int_coord = int_coord._apply_diffrot(rotatedsun_coord.duration,
                                             rotatedsun_coord.rotation_model)

        # Transform from HGS
        return int_coord.transform_to(hgs_frame)  # obstime change handled here

    _rotatedsun_cache[framecls] = _RotatedSunFramecls
    return _RotatedSunFramecls


@add_common_docstring(**_variables_for_parse_time_docstring())
class RotatedSunFrame(SunPyBaseCoordinateFrame):
    """
    A frame that applies solar rotation to a base coordinate frame.

    .. note::

        See :ref:`sunpy-topic-guide-coordinates-rotatedsunframe` for how to use this class.

    In essence, the coordinate axes of the frame are distorted by differential solar rotation.
    This allows using a coordinate representation at one time (at the ``obstime`` of the base
    coordinate frame) to point to a location at a different time that has been differentially
    rotated by the time difference (``duration``).

    Parameters
    ----------
    representation : `~astropy.coordinates.BaseRepresentation` or ``None``
        A representation object or ``None`` to have no data. Alternatively, use coordinate
        component keyword arguments, which depend on the base frame.
    base : `~astropy.coordinates.SkyCoord` or low-level coordinate object.
        The coordinate which specifies the base coordinate frame. The frame must be a SunPy frame.
    duration : `~astropy.units.Quantity` or `~astropy.time.TimeDelta`
        The duration of solar rotation (defaults to zero days).
    rotated_time : {parse_time_types}
        The time to rotate the Sun to. If provided, ``duration`` will be set to the difference
        between this time and the observation time in ``base``.
    rotation_model : `str`
        Accepted model names are ``'howard'`` (default), ``'snodgrass'``, ``'allen'``, and ``'rigid'``.
        See the documentation for :func:`~sunpy.physics.differential_rotation.differential_rotate` for differences
        between these models.

    Notes
    -----
    ``RotatedSunFrame`` is a factory class. That is, the objects that it
    yields are *not* actually objects of class ``RotatedSunFrame``. Instead,
    distinct classes are created on-the-fly for whatever the frame class is
    of ``base``.
    """
    # This code reuses significant code from Astropy's implementation of SkyOffsetFrame
    # See licenses/ASTROPY.rst

    # We don't want to inherit the frame attributes of SunPyBaseCoordinateFrame (namely `obstime`)
    # Note that this does not work for Astropy 4.3+, so we need to manually remove it below
    _inherit_descriptors_ = False

    # Even though the frame attribute `base` is a coordinate frame, we use `Attribute` instead of
    # `CoordinateAttribute` because we are preserving the supplied frame rather than converting to
    # a common frame.
    base = Attribute()

    duration = QuantityAttribute(default=0*u.day)
    rotation_model = Attribute(default='howard')

    def __new__(cls, *args, **kwargs):
        # We don't want to call this method if we've already set up
        # a rotated-Sun frame for this class.
        if not (issubclass(cls, RotatedSunFrame) and cls is not RotatedSunFrame):
            # We get the base argument, and handle it here.
            base_frame = kwargs.get('base', None)
            if base_frame is None:
                raise TypeError("Can't initialize a RotatedSunFrame without a `base` keyword.")

            # If a SkyCoord is provided, use the underlying frame
            if hasattr(base_frame, 'frame'):
                base_frame = base_frame.frame

            newcls = _make_rotatedsun_cls(base_frame.__class__)
            return newcls.__new__(newcls, *args, **kwargs)

        # https://stackoverflow.com/questions/19277399/why-does-object-new-work-differently-in-these-three-cases
        # See above for why this is necessary. Basically, because some child
        # may override __new__, we must override it here to never pass
        # arguments to the object.__new__ method.
        if super().__new__ is object.__new__:
            return super().__new__(cls)
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        # Validate inputs
        base_frame = kwargs.get('base')
        if base_frame is None or base_frame.obstime is None:
            raise ValueError("The base coordinate frame must have a defined `obstime`.")

        duration = kwargs.pop('duration', None)
        rotated_time = kwargs.pop('rotated_time', None)

        if duration is not None and rotated_time is not None:
            raise ValueError("Specify either `duration` or `rotated_time`, not both.")

        if duration is not None:
            rotated_time = base_frame.obstime + duration

        if rotated_time is not None:
            kwargs['duration'] = (parse_time(rotated_time) - base_frame.obstime).to('day')

        super().__init__(*args, **kwargs)

        # Move data out from the base frame
        if self.base.has_data:
            if not self.has_data:
                # If the duration is an array but the data is scalar, upgrade data to an array
                if self.base.data.isscalar and not self.duration.isscalar:
                    self._data = self.base.data._apply('repeat', self.duration.shape)
                else:
                    self._data = self.base.data
            self._base = self.base.replicate_without_data()

    def as_base(self):
        """
        Returns a coordinate with the current representation and in the base coordinate frame.

        This method can be thought of as "removing" the
        `~sunpy.coordinates.metaframes.RotatedSunFrame` layer. Be aware that this method is not
        merely a coordinate transformation, because this method changes the location in inertial
        space that is being pointed to.
        """
        return self.base.realize_frame(self.data)

    @property
    def rotated_time(self):
        """
        Returns the sum of the base frame's observation time and the rotation of duration.
        """
        return self.base.obstime + self.duration

    def __reduce__(self):
        return (_rotatedsunframe_reducer, (self.base,), self.__dict__)


def _rotatedsunframe_reducer(base):
    return RotatedSunFrame.__new__(RotatedSunFrame, base=base)


# For Astropy 4.3+, we need to manually remove the `obstime` frame attribute from RotatedSunFrame
if 'obstime' in RotatedSunFrame.frame_attributes:
    del RotatedSunFrame.frame_attributes['obstime']



from sunpy.net._attrs import Detector, Time, Wavelength
from sunpy.net.attr import AttrAnd, AttrWalker, DataAttr, SimpleAttr

__all__ = ["Dataset", "Limit" , "Target" , "Tags"]

class Dataset(SimpleAttr):
    """
    Name of the dataset required.
    """

class Limit(SimpleAttr):
    """
    Sets the max number of results, defaults to 20.
    """

class Target(SimpleAttr):
    """
    Indicates the observation region. Flare site used when flare flag is set.
    Source of information observation planning database, or telemetry if flare flag is set.
    some of the targets include
    Active Region (AH)
    Coronal Hole (CH)
    Flare (FS)
    Quiet Region (QR)
    """

class Tags(SimpleAttr):
    """
    a simple information associated to a solar observation,
    for example "moon transit"  etc.
    To each metadata instance can be associated 0 or more tags.
    """

walker = AttrWalker()

@walker.add_creator(AttrAnd, DataAttr)
def _create(wlk, query):
    map_ = {}
    wlk.apply(query, map_)
    return [map_]

@walker.add_applier(AttrAnd)
def _apply(wlk, query, imap):
    for iattr in query.attrs:
        wlk.apply(iattr, imap)


@walker.add_applier(Wavelength)
def _apply(wlk,query,imap):
    imap["wavelnth"] = int(query.min.value)


@walker.add_applier(Dataset)
def _apply(wlk,query,imap):
    imap["datasets"] = f"metadata_{query.value}"


@walker.add_applier(Time)
def _apply(wlk,query,imap):
    imap["date_end__gte"] = query.start.value
    imap['data_beg__lte'] = query.end.value


@walker.add_applier(Detector)
def _apply(wlk,query,imap):
    imap["detector__iexact"] = query.value


@walker.add_applier(Limit)
def _apply(wlk,query,imap):
    imap["limit"] = int(query.value)

@walker.add_applier(Target)
def _apply(wlk,query,imap):
    imap["target__in"] = query.value

@walker.add_applier(Tags)
def _apply(wlk,query,imap):
    imap["tag"] = query.value



"""
This package contains all of SunPy's test data.
"""
import os
import re
import glob
import fnmatch
from pathlib import Path

import numpy as np

import astropy.io.fits
from astropy.utils.data import get_pkg_data_filename

import sunpy
import sunpy.io._fits as _fits

__all__ = [
    'rootdir',
    'file_list',
    'get_test_filepath',
    'get_test_data_filenames',
    'get_dummy_map_from_header',
    'write_header_file_from_image_file',
]

rootdir = Path(os.path.dirname(sunpy.__file__)) / "data" / "test"
file_list = glob.glob(os.path.join(rootdir, '*.[!p]*'))


def get_test_filepath(filename, package="sunpy.data.test", **kwargs):
    """
    Return the full path to a test file in the ``data/test`` directory.

    Parameters
    ----------
    filename : `str`
        The name of the file inside the ``data/test`` directory.
    package : `str`, optional
        The package in which to look for the file. Defaults to "sunpy.data.test".

    Returns
    -------
    filepath : `str`
        The full path to the file.

    Notes
    -----
    This is a wrapper around `astropy.utils.data.get_pkg_data_filename` which
    sets the ``package`` kwarg to be 'sunpy.data.test`.
    """
    if isinstance(filename, Path):
        # NOTE: get_pkg_data_filename does not accept Path objects
        filename = filename.as_posix()
    return get_pkg_data_filename(filename, package=package, **kwargs)


def get_test_data_filenames():
    """
    Return a list of all test files in ``data/test`` directory.

    This ignores any ``py``, ``pyc`` and ``__*__`` files in these directories.

    Returns
    -------
    `list`
        The name of all test files in ``data/test`` directory.
    """
    get_test_data_filenames_list = []
    excludes = ['*.pyc', '*'+os.path.sep+'__*__', '*.py']
    excludes = r'|'.join([fnmatch.translate(x) for x in excludes]) or r'$.'

    for root, _, files in os.walk(rootdir):
        files = [Path(root) / f for f in files]
        files = [f for f in files if not re.match(excludes, str(f))]
        get_test_data_filenames_list.extend(files)

    return get_test_data_filenames_list


def write_image_file_from_header_file(header_file, fits_directory):
    """
    Given a header-only file ``header_file``, write a dummy FITS file
    with the same name to ``fits_directory``

    Parameters
    ----------
    header_file: `~pathlib.Path`
    fits_directory: `~pathlib.Path`

    Returns
    -------
    fits_file: `~pathlib.Path`
        Path to dummy FITS file
    """
    header = astropy.io.fits.Header.fromtextfile(header_file)
    shape = [header[f'naxis{i+1}'] for i in range(header['naxis'])]
    data = np.random.rand(*shape[::-1])
    if 'BITPIX' in header:
        data = data.astype(astropy.io.fits.BITPIX2DTYPE[header['BITPIX']])
    hdu = astropy.io.fits.PrimaryHDU(data=data, header=header, do_not_scale_image_data=True, scale_back=True)
    fits_file = os.fspath(fits_directory.joinpath(header_file.with_suffix('.fits').name))
    hdu.writeto(fits_file)
    return fits_file


def get_header_from_file(filename, package="sunpy.data.test"):
    """
    Return a Header from the test file.
    """
    filepath = get_test_filepath(filename, package=package)
    return _fits.format_comments_and_history(astropy.io.fits.Header.fromtextfile(filepath))


def get_dummy_map_from_header_obj(header):
    data = np.random.rand(header['NAXIS2'], header['NAXIS1'])
    if 'BITPIX' in header:
        data = data.astype(astropy.io.fits.BITPIX2DTYPE[header['BITPIX']])
    # NOTE: by reading straight from the data header pair, we are skipping
    # the fixes that are applied in sunpy.io._fits, e.g. the waveunit fix
    # Would it be better to write this whole map back to a temporary file and then
    # read it back in by passing in the filename instead?
    return sunpy.map.Map(data, header)


def get_dummy_map_from_header(filename, package="sunpy.data.test"):
    """
    Generate a dummy `~sunpy.map.Map` from header-only test data.

    The "image" will be random numbers with the correct shape
    as specified by the header.
    """
    import sunpy.map

    header = get_header_from_file(filename, package)
    return get_dummy_map_from_header_obj(header)


def write_header_file_from_image_file(image_path, header_path, hdu=0, **kwargs):
    """
    Convert a FITS file containing an image and a header
    to a plaintext file containing only the string representation
    of the header.

    This is to be used in the context of generating header-only test data
    files.

    Parameters
    -----------
    image_path : path-like
        Path to original image data FITS file
    header_path : path-like
        Path to header-only test data plaintext file
    hdu : `int`, optional
        HDU index of the header to write the header for

    Additional Parameters
    ---------------------
    Keyword arguments accepted by `~astropy.io.fits.Header.totextfile`

    See Also
    --------
    get_dummy_map_from_header
    """
    with astropy.io.fits.open(image_path) as hdul:
        hdul[hdu].header.totextfile(header_path, **kwargs)



"""
====================
HMI Showcase: Cutout
====================

This example demonstrates how to plot a cutout region of a `~sunpy.map.Map`
with connector lines that indicate the region of interest in the full-disk
image.

Since this example deals with the creation of a specific style of image, there
are multiple lines that deal directly with matplotlib axes.
"""
import matplotlib.colors
import matplotlib.pyplot as plt
from matplotlib.patches import ConnectionPatch

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.data.sample import HMI_LOS_IMAGE

##############################################################################
# First, we use the sample HMI LOS image and focus the cutout over an active
# region near the solar center.

magnetogram = sunpy.map.Map(HMI_LOS_IMAGE).rotate()
left_corner = SkyCoord(Tx=-142*u.arcsec, Ty=50*u.arcsec, frame=magnetogram.coordinate_frame)
right_corner = SkyCoord(Tx=158*u.arcsec, Ty=350*u.arcsec, frame=magnetogram.coordinate_frame)

##############################################################################
# We clean up the magnetogram by masking off all data that is beyond the solar
# limb.

hpc_coords = sunpy.map.all_coordinates_from_map(magnetogram)
mask = ~sunpy.map.coordinate_is_on_solar_disk(hpc_coords)
magnetogram_big = sunpy.map.Map(magnetogram.data, magnetogram.meta, mask=mask)

##############################################################################
# We create the figure in two stages.
# The first stage is plotting the full-disk magnetogram.

# sphinx_gallery_defer_figures

fig = plt.figure(figsize=(7.2, 4.8))

################################################################################
# We create a nice normalization range for the image.

# sphinx_gallery_defer_figures

norm = matplotlib.colors.SymLogNorm(50, vmin=-7.5e2, vmax=7.5e2)

################################################################################
# Plot the full-disk magnetogram.

# sphinx_gallery_defer_figures

ax1 = fig.add_subplot(121, projection=magnetogram_big)
magnetogram_big.plot(axes=ax1, cmap='RdBu_r', norm=norm, annotate=False,)
magnetogram_big.draw_grid(axes=ax1, color='black', alpha=0.25, lw=0.5)

################################################################################
# These lines deal with hiding the axis, its ticks and labels.

# sphinx_gallery_defer_figures

for coord in ax1.coords:
    coord.frame.set_linewidth(0)
    coord.set_ticks_visible(False)
    coord.set_ticklabel_visible(False)

################################################################################
# We draw the rectangle around the region we plan to showcase in the cutout image.

# sphinx_gallery_defer_figures

magnetogram_big.draw_quadrangle(left_corner, top_right=right_corner, edgecolor='black', lw=1)

################################################################################
# The second stage is plotting the zoomed-in magnetogram.

# sphinx_gallery_defer_figures

magnetogram_small = magnetogram.submap(left_corner, top_right=right_corner)
ax2 = fig.add_subplot(122, projection=magnetogram_small)
im = magnetogram_small.plot(axes=ax2, norm=norm, cmap='RdBu_r', annotate=False,)
ax2.grid(alpha=0)

################################################################################
# Unlike the full-disk image, here we just clean up the axis labels and ticks.

# sphinx_gallery_defer_figures

lon, lat = ax2.coords[0], ax2.coords[1]
lon.frame.set_linewidth(1)
lat.frame.set_linewidth(1)
lon.set_axislabel('Helioprojective Longitude',)
lon.set_ticks_position('b')
lat.set_axislabel('Helioprojective Latitude',)
lat.set_axislabel_position('r')
lat.set_ticks_position('r')
lat.set_ticklabel_position('r')

################################################################################
# Now for the finishing touches, we add two lines that will connect
# the two images as well as a colorbar.

# sphinx_gallery_defer_figures

xpix, ypix = magnetogram_big.wcs.world_to_pixel(right_corner)
con1 = ConnectionPatch(
    (0, 1), (xpix, ypix), 'axes fraction', 'data', axesA=ax2, axesB=ax1,
    arrowstyle='-', color='black', lw=1
)
xpix, ypix = magnetogram_big.wcs.world_to_pixel(
    SkyCoord(right_corner.Tx, left_corner.Ty, frame=magnetogram_big.coordinate_frame))
con2 = ConnectionPatch(
    (0, 0), (xpix, ypix), 'axes fraction', 'data', axesA=ax2, axesB=ax1,
    arrowstyle='-', color='black', lw=1
)
ax2.add_artist(con1)
ax2.add_artist(con2)

pos = ax2.get_position().get_points()
cax = fig.add_axes([
    pos[0, 0], pos[1, 1]+0.01, pos[1, 0]-pos[0, 0], 0.025
])
cbar = fig.colorbar(im, cax=cax, orientation='horizontal')

################################################################################
# For the colorbar we want it to have three fixed ticks.

cbar.locator = matplotlib.ticker.FixedLocator([-1e2, 0, 1e2])
cbar.set_label("LOS Magnetic Field [gauss]", labelpad=-40, rotation=0)
cbar.update_ticks()
cbar.ax.xaxis.set_ticks_position('top')

plt.show()



"""
Solar Orbiter Map subclass definitions.
"""
import astropy.units as u
from astropy.coordinates import CartesianRepresentation
from astropy.visualization import AsinhStretch, ImageNormalize

from sunpy.coordinates import HeliocentricInertial
from sunpy.map import GenericMap
from sunpy.map.sources.source_type import source_stretch

__all__ = ['EUIMap']


class EUIMap(GenericMap):
    """
    EUI Image Map

    The Extreme Ultraviolet Imager (EUI) is a remote sensing instrument onboard the
    Solar Orbiter (SolO) spacecraft. EUI has three telescopes that image the Sun in
    Lyman-alpha (1216 Å) and the EUV (174 Å and 304 Å). The three telescopes are the
    Full Sun Imager (FSI) and two High Resolution Imagers (HRI). The FSI images the
    whole Sun in both 174 Å and 304 Å. The EUV and Lyman-alpha HRI telescopes image a
    1000"-by-1000" patch in 174 Å and 1216 Å, respectively.

    References
    ----------
    * `Solar Orbiter Mission Page <https://sci.esa.int/web/solar-orbiter/>`__
    * `EUI Instrument Page <https://www.sidc.be/EUI/about/instrument>`__
    * Instrument Paper: :cite:t:`rochus_solar_2020`
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        self._nickname = self.detector
        self.plot_settings['norm'] = ImageNormalize(
            stretch=source_stretch(self.meta, AsinhStretch(0.01)), clip=False)

    @property
    def _rotation_matrix_from_crota(self):
        return super()._rotation_matrix_from_crota(crota_key='CROTA')

    @property
    def processing_level(self):
        if self.meta.get('level'):
            # The level number is prepended by the letter L
            return int(self.meta.get('level')[1:])

    @property
    def waveunit(self):
        # EUI JP2000 files do not have the WAVEUNIT key in the metadata.
        # However, the FITS files do.
        # The EUI metadata spec says the WAVELNTH key is always expressed
        # in Angstroms so we assume this if the WAVEUNIT is missing.
        return super().waveunit or u.Angstrom

    @property
    def _supported_observer_coordinates(self):
        return [(('hcix_obs', 'hciy_obs', 'hciz_obs'),
                 {'x': self.meta.get('hcix_obs'),
                  'y': self.meta.get('hciy_obs'),
                  'z': self.meta.get('hciz_obs'),
                  'unit': u.m,
                  'representation_type': CartesianRepresentation,
                  'frame': HeliocentricInertial})] + super()._supported_observer_coordinates

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an EUI image"""
        is_solo = 'solar orbiter' in str(header.get('obsrvtry', '')).lower()
        is_eui = str(header.get('instrume', '')).startswith('EUI')
        return is_solo and is_eui



import warnings

import numpy as np
import pytest

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.io.fits.verify import VerifyWarning
from astropy.time import Time

import sunpy.coordinates
import sunpy.map
import sunpy.sun
from sunpy.coordinates import sun
from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.io._file_tools import read_file


@pytest.fixture
def test_map(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def hmi_test_map():
    (data, header), = read_file(get_test_filepath('resampled_hmi.fits'))

    # Get rid of the blank keyword to prevent some astropy fits fixing warnings
    header.pop('BLANK')
    header.pop('CRDER2')
    header.pop('CRDER1')
    return sunpy.map.Map((data, header))


@pytest.fixture
def aia171_test_map():
    (data, header), = read_file(get_test_filepath('aia_171_level1.fits'))

    # Get rid of the blank keyword to prevent some astropy fits fixing warnings
    header.pop('BLANK')
    return sunpy.map.Map((data, header))


@pytest.fixture
def adjusted_test_maps(aia171_test_map, hmi_test_map):
    # The test maps have wildly different observation times, which throws off compositing
    hmi_test_map.meta['date-obs'] = aia171_test_map.meta['date-obs']
    hmi_test_map.meta['t_obs'] = aia171_test_map.meta['t_obs']
    # Also set the HMI observer location to be the same as the AIA observer location
    del hmi_test_map.meta['crln_obs']
    del hmi_test_map.meta['crlt_obs']
    hmi_test_map.meta['hgln_obs'] = aia171_test_map.observer_coordinate.lon.to_value('deg')
    hmi_test_map.meta['hglt_obs'] = aia171_test_map.observer_coordinate.lat.to_value('deg')
    return aia171_test_map, hmi_test_map


@pytest.fixture
def aia171_roll_map(aia171_test_map):
    return aia171_test_map.rotate(-45*u.deg)


@pytest.fixture
def heliographic_test_map():
    (data, header), = read_file(get_test_filepath('heliographic_phase_map.fits.gz'))

    # Fix unit strings to prevent some astropy fits fixing warnings
    header['CUNIT1'] = 'deg'
    header['CUNIT2'] = 'deg'
    # Set observer location to avoid warnings later
    header['HGLN_OBS'] = 0.0
    return sunpy.map.Map((data, header))


@pytest.fixture
def aia171_test_map_with_mask(aia171_test_map):
    shape = aia171_test_map.data.shape
    mask = np.zeros_like(aia171_test_map.data, dtype=bool)
    mask[0:shape[0] // 2, 0:shape[1] // 2] = True
    return sunpy.map.Map(np.ma.array(aia171_test_map.data, mask=mask), aia171_test_map.meta)


@pytest.fixture
def generic_map():
    data = np.arange(36, dtype=np.float64).reshape((6, 6))
    dobs = Time('1970-01-01T00:00:00')
    l0 = sun.L0(dobs).to_value(u.deg)
    b0 = sun.B0(dobs).to_value(u.deg)
    dsun = sun.earth_distance(dobs).to_value(u.m)
    header = {
        'CRVAL1': 0,
        'CRVAL2': 0,
        'CRPIX1': 5,
        'CRPIX2': 5,
        'CDELT1': 10,
        'CDELT2': 10,
        'CUNIT1': 'arcsec',
        'CUNIT2': 'arcsec',
        'CTYPE1': 'HPLN-TAN',
        'CTYPE2': 'HPLT-TAN',
        'PC1_1': 0,
        'PC1_2': -1,
        'PC2_1': 1,
        'PC2_2': 0,
        'NAXIS1': 6,
        'NAXIS2': 6,
        'date-obs': dobs.isot,
        'crln_obs': l0,
        'crlt_obs': b0,
        "dsun_obs": dsun,
        'mjd-obs': 40587.0,
        'obsrvtry': 'Foo',
        'detector': 'bar',
        'wavelnth': 10,
        'waveunit': 'm',
        'bunit': 'DN/s',
    }
    return sunpy.map.Map((data, header))


@pytest.fixture
def simple_map():
    """
    A simple 9x9 map, with its center at (0, 0),
    and scaled differently in each direction.
    """
    data = np.arange(81).reshape((9, 9))
    ref_coord = SkyCoord(0.0, 0.0, frame='helioprojective', obstime='2020-01-01 00:00:00', unit='deg',
                         observer=SkyCoord(0 * u.deg, 0 * u.deg, 1 * u.AU, frame='heliographic_stonyhurst'))
    ref_pix = [4, 4] * u.pix
    scale = [2, 1] * u.arcsec / u.pix
    header = sunpy.map.make_fitswcs_header(data, ref_coord, reference_pixel=ref_pix, scale=scale)
    return sunpy.map.Map(data, header)


@pytest.fixture
def carrington_map():
    # This is a 20 x 20 map in a Carrington frame, with the reference pixel *not* at the
    # equator. This results in a non-linear transformation between pixel and world
    # coordinates, so is ideal for testing situations where the non-linearity matters
    data = np.arange(20**2).reshape((20, 20))
    obstime = '2020-01-01'
    observer = SkyCoord(0*u.deg, 0*u.deg, frame='heliographic_stonyhurst', obstime=obstime)
    ref_coord = SkyCoord(120*u.deg, -70*u.deg, frame='heliographic_carrington', obstime=obstime, observer=observer)
    ref_pix = [0, 0] * u.pix
    scale = [2, 1] * u.deg / u.pix
    header = sunpy.map.make_fitswcs_header(data, ref_coord, reference_pixel=ref_pix, scale=scale)
    return sunpy.map.Map(data, header)


@pytest.fixture
def eit_test_map():
    """
    Load SunPy's test EIT image.
    """
    return get_dummy_map_from_header(get_test_filepath("EIT_header/efz20040301.020010_s.header"))


@pytest.fixture
def sample_171():
    from sunpy.data.sample import AIA_171_IMAGE

    # Need to bypass
    # VerifyWarning: Invalid 'BLANK' keyword in header.
    # The 'BLANK' keyword is only applicable to integer data, and will be ignored in this HDU.
    with fits.open(AIA_171_IMAGE) as hdu:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=VerifyWarning)
            return sunpy.map.Map(hdu[1].data, hdu[1].header)


@pytest.fixture
def sample_hmi():
    from sunpy.data.sample import HMI_LOS_IMAGE

    return sunpy.map.Map(HMI_LOS_IMAGE)



"""
SUIT map subclass definitions
"""
from astropy.visualization import AsinhStretch, ImageNormalize

from sunpy.map.mapbase import GenericMap
from sunpy.map.sources.source_type import source_stretch

__all__ = ["SUITMap"]

class SUITMap(GenericMap):
    """
    SUIT Image Map.

    The Solar Ultraviolet Imaging Telescope (SUIT) is one of the remote sensing
    payloads on board the Aditya-L1 mission of the Indian Space Research Organization (ISRO)
    that was launched on September 2, 2023 and located at the first Lagrange point.
    SUIT is designed to provide near-simultaneous full-disk
    and region-of-interest images of the Sun at various heights, slicing through the photosphere
    and chromosphere, employing an array of 11 scientifically calibrated filters (3 broad-band and 8 narrow-band)
    strategically positioned within the wavelength range of 200 to 400 nanometers.

    The SUIT data is available to download from the PRADAN interface of ISRO.

    References
    ----------
    * `Home Page: <https://suit.iucaa.in/>`__
    * `Mission Page: <https://suit.iucaa.in/about_SUIT>`__
    * `Publications Page: <https://suit.iucaa.in/node/5>`__
    * `Data Download PRADAN Page: <https://pradan.issdc.gov.in/al1>`__
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        self._nickname = self.detector
        filtername = header.get("FTR_NAME", "").strip()
        filternorms = {
            "NB01": 0.1,
            "NB02": 9.0,
            "NB03": 0.25,
            "NB04": 0.2,
            "NB05": 2.0,
            "NB06": 0.25,
            "NB07": 0.25,
            "NB08": 0.1,
            "BB01": 0.8,
            "BB02": 0.4,
            "BB03": 0.8,
        }
        self.plot_settings["cmap"] = f"suit_{filtername.lower()}"
        self.plot_settings["norm"] = ImageNormalize(stretch=source_stretch(self.meta, AsinhStretch(filternorms.get(filtername, 0.2))), clip=False)

    @property
    def observatory(self):
        return self.meta.get("MISSION", "Aditya-L1")

    @property
    def instrument(self):
        return "SUIT"

    @property
    def detector(self):
        return self.meta.get("PNAME", "SUIT")

    @property
    def processing_level(self):
        return self.meta.get("F_LEVEL", "")

    @property
    def _date_obs(self):
        """
        Time of observation
        """
        return self._get_date("T_OBS")

    @property
    def reference_date(self):
        """
        The reference time is the time of Shutter open time. Does not include the exposure time.
        """
        return self._get_date("T_OBS") or super().reference_date

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """
        Determines if header corresponds to a SUIT image
        """
        return header.get("PNAME", "").strip().lower() == "suit"



"""
Script to render to terminal, the changelog entry for the latest version in markdown so it can be copied and pasted for a GitHub release.

It requires pandoc to be installed on your system.
"""
import subprocess
from pathlib import Path

res = subprocess.run(['pandoc', '--wrap=none', '-t', 'gfm', str(Path(__file__).parent.parent / "CHANGELOG.rst")], capture_output=True)
print(res.stdout.decode('ascii').split('\n# ')[0])



"""
===============================================
Tracking an Active Region Across the Solar Disk
===============================================

This example demonstrates how to track an active region as it rotates across the solar disk
and make cutouts around that active region at each time step to build a tracked datacube.
"""
# sphinx_gallery_thumbnail_number = 3
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.visualization import ImageNormalize, SqrtStretch

import sunpy.map
from sunpy.coordinates import propagate_with_solar_surface
from sunpy.net import Fido
from sunpy.net import attrs as a

###############################################################################
# First, let's download a series of images in time using `~sunpy.net.Fido`.
# In this example, we will download a series of AIA 171 Å images observed over the course
# of half of a day at a cadence of 1 image every 1 hour.

query = Fido.search(a.Time('2018-05-30 00:00:00', '2018-05-30 12:00:00'),
                    a.Instrument.aia,
                    a.Wavelength(171*u.angstrom),
                    a.Sample(1*u.h))
print(query)
files = Fido.fetch(query)

###############################################################################
# Now that we have a set of images in time, we can create a `~sunpy.map.MapSequence` to hold all of them
# and animate that sequence in time.

aia_sequence = sunpy.map.Map(files, sequence=True)

fig = plt.figure()
ax = fig.add_subplot(projection=aia_sequence[0])
norm = norm=ImageNormalize(vmin=0, vmax=3e3, stretch=SqrtStretch())
ani = aia_sequence.plot(axes=ax, norm=norm)

###############################################################################
# Next, let's crop one of the maps in our sequence to the active region of interest.

corner = SkyCoord(Tx=-250*u.arcsec, Ty=0*u.arcsec, frame=aia_sequence[6].coordinate_frame)
cutout_map = aia_sequence[6].submap(corner, width=500*u.arcsec, height=500*u.arcsec)

fig = plt.figure()
ax = fig.add_subplot(projection=cutout_map)
cutout_map.plot(axes=ax)

###############################################################################
# Now that we have our cutout around our active region, we can reproject each map in our sequence to
# the WCS of that cutout. Additionally, we will use the `~sunpy.coordinates.propagate_with_solar_surface`
# context manager to adjust the field of view of the cutout with the rotation of the solar surface.
# We use the ``preserve_date_obs`` keyword argument to preserve the original observation time in each
# reprojected map. Otherwise, the observation time of the reprojected maps would all be the same as the
# cutout WCS. Note that using this keyword does not affect the resulting coordinate frame of the reprojected
# map which will be defined by the date of the cutout WCS.

with propagate_with_solar_surface():
    aia_sequence_aligned = sunpy.map.Map(
        [m.reproject_to(cutout_map.wcs, preserve_date_obs=True) for m in aia_sequence],
        sequence=True
    )

###############################################################################
# Finally, we can animate our sequence of reprojected cutouts to confirm that we've tracked our active
# region of interest as a function of time across the solar disk.

fig = plt.figure()
ax = fig.add_subplot(projection=aia_sequence_aligned[0])
ani = aia_sequence_aligned.plot(axes=ax, cmap='sdoaia171', norm=norm)

plt.show()



"""
Test cases for Yohkoh SXTMap subclass
"""
import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.sources.yohkoh import SXTMap
from sunpy.util.exceptions import SunpyMetadataWarning
from .helpers import _test_private_date_setters


@pytest.fixture
def sxt_map():
    with pytest.warns(SunpyMetadataWarning, match='Missing CTYPE'):
        return get_dummy_map_from_header(get_test_filepath("YohkohSXT.header"))


def test_fits_to_sxt(sxt_map):
    """Tests the creation of SXTMap using FITS."""
    assert isinstance(sxt_map, SXTMap)


def test_is_datasource_for(sxt_map):
    """Test the is_datasource_for method of SXTMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert sxt_map.is_datasource_for(sxt_map.data, sxt_map.meta)


def test_reference_date(sxt_map):
    assert sxt_map.reference_date.isot == "1991-11-05T11:10:24.018"


def test_date(sxt_map):
    assert sxt_map.date.isot == "1991-11-05T11:10:24.018"


def test_private_date_setters(sxt_map):
    _test_private_date_setters(sxt_map)


def test_observatory(sxt_map):
    """Tests the observatory property of the SXTMap object."""
    assert sxt_map.observatory == "Yohkoh"


def test_detector(sxt_map):
    """Tests the detector property of the SXTMap object"""
    assert sxt_map.detector == "SXT"


def test_measurement(sxt_map):
    """Tests the measurement property of the SXTMap object."""
    assert sxt_map.measurement == 'Al01'


def test_wavelength(sxt_map):
    """Tests that the wavelength of the SXTMap is always None"""
    assert sxt_map.wavelength is None


def test_heliographic_longitude(sxt_map):
    assert u.allclose(sxt_map.heliographic_longitude, 0.002357818089886795 * u.deg)


def test_heliographic_latitude(sxt_map):
    assert u.allclose(sxt_map.heliographic_latitude, 3.955251457847305 * u.deg)


def test_dsun(sxt_map):
    assert u.allclose(sxt_map.dsun, 148328341.63429108 * u.km)


def test_wcs(sxt_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    with pytest.warns(SunpyMetadataWarning, match='Missing CTYPE'):
        sxt_map.pixel_to_world(0*u.pix, 0*u.pix)



"""
Functions for geometrical image transformation and warping.
"""
import sys
import time
import numbers
from functools import wraps
from collections import namedtuple

import numpy as np
import scipy.ndimage
from scipy.signal import convolve2d

from sunpy import log
from sunpy.util.exceptions import warn_user

__all__ = ['add_rotation_function', 'affine_transform']


def affine_transform(image, rmatrix, order=3, scale=1.0, image_center=None,
                     recenter=False, missing=np.nan, *, method='scipy', clip=True):
    """
    Rotates, shifts and scales an image.

    This function supports NaN values in the input image and supports using NaN for
    pixels in the output image that are beyond the extent of the input image.
    Handling NaN values in the input image requires additional computation time.

    Parameters
    ----------
    image : `numpy.ndarray`
        2D image to be rotated.
    rmatrix : `numpy.ndarray` that is 2x2
        Linear transformation rotation matrix.
    order : `int` 0-5, optional
        Interpolation order to be used, defaults to 3. The precise meaning depends
        on the rotation method specified by ``method``.
    scale : `float`
        A scale factor for the image with the default being no scaling.
    image_center : tuple, optional
        The point in the image to rotate around (axis of rotation).
        Defaults to the center of the array.
    recenter : `bool` or array-like, optional
        Move the axis of rotation to the center of the array or recenter coords.
        Defaults to `False`.
    missing : `float`, optional
        The value to use for pixels in the output image that are beyond the extent
        of the input image.
        Defaults to `numpy.nan`.
    method : {{{rotation_function_names}}}, optional
        Rotation function to use.
        Defaults to ``'scipy'``.
    clip : `bool`, optional
        If `True`, clips the pixel values of the output image to the range of the
        input image (including the value of ``missing``, if used).
        Defaults to `True`.

    Returns
    -------
    `numpy.ndarray`:
        New rotated, scaled and translated image.

    Notes
    -----
    For each NaN pixel in the input image, one or more pixels in the output image
    will be set to NaN, with the size of the pixel region affected depending on the
    interpolation order. All currently implemented rotation methods require a
    convolution step to handle image NaNs. This convolution normally uses
    :func:`scipy.signal.convolve2d`, but if `OpenCV <https://opencv.org>`__ is
    installed, the faster |cv2_filter2D|_ is used instead.

    See :func:`~sunpy.image.transform.add_rotation_function` for how to add a
    different rotation function.
    """
    rmatrix = rmatrix / scale
    array_center = (np.array(image.shape)[::-1] - 1) / 2.0

    # Make sure the image center is an array and is where it's supposed to be
    if image_center is not None:
        image_center = np.asanyarray(image_center)
    else:
        image_center = array_center

    # Determine center of rotation based on use (or not) of the recenter keyword
    if recenter:
        rot_center = array_center
    else:
        rot_center = image_center

    displacement = np.dot(rmatrix, rot_center)
    shift = image_center - displacement

    method = _get_transform_method(method)

    # Transform the image using the appropriate function
    rotated_image = _rotation_registry[method].function(image, rmatrix, shift, order, missing, clip)

    return rotated_image


def _get_transform_method(method):
    # This is re-used in affine_transform and GenericMap.rotate
    if method not in _rotation_registry:
        raise ValueError(f'Method {method} not in supported methods: '
                         f'{_rotation_registry.keys()}')
    return method


def add_rotation_function(name, *, allowed_orders,
                          handles_clipping, handles_image_nans, handles_nan_missing):
    """
    Decorator to add a rotation function to the registry of selectable
    implementations.

    Each registered rotation function becomes a selectable option for
    :func:`sunpy.image.transform.affine_transform` and
    :meth:`sunpy.map.GenericMap.rotate`. Those two routines are required to handle
    clipping the output image, NaNs in the input image, and NaN as the value to use
    for pixels in the output image that are beyond the extent of the input image.
    If the supplied rotation function cannot provide one or more of these capabilities,
    the decorator is able to provide them instead.

    The decorator requires the parameters listed under ``Parameters``. The decorated
    rotation function must accept the parameters listed under ``Other Parameters``
    in that order and return the rotated image.

    Parameters
    ----------
    name : `str`
        The name that will be used to select the rotation function
    allowed_orders : `set`
        The allowed values for the ``order`` parameter.
    handles_clipping : `bool`
        Specifies whether the rotation function will internally perform clipping.
        If ``False``, the rotation function will always receive ``False`` for the
        ``clip`` input parameter.
    handles_image_nans : `bool`
        Specifies whether the rotation function will internally handle NaNs in the
        input image. If ``False``, the rotation function is guaranteed to be
        provided an image without any NaNs.
    handles_nan_missing : `bool`
        Specifies whether the rotation function will internally handle NaN as the
        ``missing`` value. If ``False``, the rotation function will never receive
        NaN, but instead receive a value in the input range of the image.

    Other Parameters
    ----------------
    image : `numpy.ndarray`
        The image, which could be integers or floats
    matrix : `numpy.ndarray` that is 2x2
        The linear transformation matrix (e.g., rotation+scale+skew)
    shift : 2-element `numpy.ndarray`
        The translational shift to apply to the image in each axis
    order : `int`
        The numerical parameter that controls the degree of interpolation
    missing : `float`
        The value to use for outside the bounds of the original image
    clip : `bool`
        Whether to clip the output image to the range of the input image

    Notes
    -----
    The docstring of the rotation function should be a bulleted list of notes
    specific to the rotation function. It will be appended to ``Notes`` section of
    the docstring for :func:`~sunpy.image.transform.affine_transform`.

    The rotation function is supplied the input image directly, so the function
    should not modify the image in place.

    Setting any of the ``handles_*`` parameters to ``False`` means that computation
    will be performed to modify the image returned by the rotation function before
    it is returned to :func:`~sunpy.image.transform.affine_transform`.

    If the decorator is handling image NaNs on behalf of the rotation function
    (i.e., ``handles_image_nans=False``), pixels in the output image will be set to
    NaN if they are within a certain neighborhood size that depends on the ``order``
    parameter. This step requires an additional image convolution, which might be
    avoidable if the rotation function were able to internally handle image NaNs.
    This convolution normally uses :func:`scipy.signal.convolve2d`, but if
    `OpenCV <https://opencv.org>`__ is installed, the faster |cv2_filter2D|_ is
    used instead.
    """
    def decorator(rotation_function):
        @wraps(rotation_function)
        def wrapper(image, matrix, shift, order, missing, clip):
            if order not in allowed_orders:
                raise ValueError(f"{order} is one of the allowed orders for method '{name}': "
                                 f"{set(allowed_orders)}")

            clip_to_use = clip if handles_clipping else False

            # Check if missing is NaN and needs to be externally handled
            needs_missing_handling = not handles_nan_missing and np.isnan(missing)

            # Check if there are any NaNs in the image that need to be externally handled
            if not handles_image_nans:
                isnan = np.isnan(image)
                needs_nan_handling = np.any(isnan)
            else:
                needs_nan_handling = False

            # If either is needed, change the NaNs to the median of the input image
            if needs_missing_handling or needs_nan_handling:
                substitute = np.nanmedian(image)
            missing_to_use = substitute if needs_missing_handling else missing
            image_to_use = np.nan_to_num(image, nan=substitute) if needs_nan_handling else image

            t = time.perf_counter()
            rotated_image = rotation_function(image_to_use, matrix, shift, order,
                                              missing_to_use, clip_to_use)
            log.debug(f"{name} rotating image: {time.perf_counter() - t:.3f} s")

            # If needed, restore the NaNs
            if needs_nan_handling:
                # Use a convolution to find all pixels that are appreciably affected by NaNs
                # The kernel size depends on the interpolation order, but are empirically defined
                # because a given pixel can affect every other pixel under spline interpolation
                sizes = [1, 1, 5, 5, 7, 7]

                t = time.perf_counter()
                try:
                    # If OpenCV is installed, its convolution function is much faster
                    import cv2
                    expanded_nans = cv2.filter2D(isnan.astype(float), -1,
                                                 np.ones((sizes[order], sizes[order])),
                                                 borderType=cv2.BORDER_CONSTANT)
                except ImportError:
                    expanded_nans = convolve2d(isnan.astype(float),
                                               np.ones((sizes[order], sizes[order])),
                                               mode='same')
                log.debug(f"{name} expanding image NaNs: {time.perf_counter() - t:.3f} s")

                t = time.perf_counter()
                rotated_nans = rotation_function(expanded_nans, matrix, shift, order=min(order, 1),
                                                 missing=0, clip=False)
                rotated_image[rotated_nans > 0] = np.nan
                log.debug(f"{name} rotating image NaNs: {time.perf_counter() - t:.3f} s")

            if needs_missing_handling:
                t = time.perf_counter()
                # Rotate a constant image to determine where to apply the NaNs for `missing`
                constant = rotation_function(np.ones_like(image), matrix, shift, order=1,
                                             missing=0, clip=False)
                rotated_image[constant < 1] = np.nan
                log.debug(f"{name} applying NaN missing: {time.perf_counter() - t:.3f} s")

            if not handles_clipping and clip and not np.all(np.isnan(rotated_image)):
                t = time.perf_counter()
                # Clip the image to the input range
                if np.isnan(missing):
                    # If `missing` is NaN, clipping to the input range is straightforward
                    rotated_image.clip(np.nanmin(image), np.nanmax(image), out=rotated_image)
                else:
                    # Otherwise, check if `missing` should be considered part of the input range
                    lower = np.nanmin([np.max([missing, np.nanmin(rotated_image)]), np.nanmin(image)])
                    upper = np.nanmax([np.min([missing, np.nanmax(rotated_image)]), np.nanmax(image)])
                    rotated_image.clip(lower, upper, out=rotated_image)
                log.debug(f"{name} clipping image: {time.perf_counter() - t:.3f} s")

            return rotated_image

        _rotation_registry[name] = _rotation_method(function=wrapper,
                                                    allowed_orders=set(allowed_orders))

        # Add the docstring of the rotation function to the docstring of affine_transform
        affine_transform.__doc__ += (f"\n    **Specific notes for the '{name}' rotation method:**"
                                     f"\n{wrapper.__doc__}")

        return wrapper
    return decorator


_rotation_method = namedtuple('_rotation_method', ['function', 'allowed_orders'])
_rotation_registry = {}


@add_rotation_function("scipy", allowed_orders=range(6),
                       handles_clipping=False, handles_image_nans=False, handles_nan_missing=True)
def _rotation_scipy(image, matrix, shift, order, missing, clip):
    """
    * Rotates using :func:`scipy.ndimage.affine_transform`
    * The ``order`` parameter is the order of the spline interpolation, and ranges
      from 0 to 5.
    * The ``mode`` parameter for :func:`~scipy.ndimage.affine_transform` is fixed to
      be ``'constant'``
    """
    rotated_image = scipy.ndimage.affine_transform(image.T, matrix, offset=shift, order=order,
                                                   mode='constant', cval=missing).T

    return rotated_image


@add_rotation_function("scikit-image", allowed_orders=range(6),
                       handles_clipping=False, handles_image_nans=False, handles_nan_missing=False)
def _rotation_skimage(image, matrix, shift, order, missing, clip):
    """
    * Rotates using :func:`skimage.transform.warp`
    * The ``order`` parameter selects from the following interpolation algorithms:

      * 0: nearest-neighbor
      * 1: bi-linear
      * 2: bi-quadratic
      * 3: bi-cubic
      * 4: bi-quartic
      * 5: bi-quintic

    * The implementation for higher orders of interpolation means that the pixels
      in the output image that are beyond the extent of the input image may not have
      exactly the value of the ``missing`` parameter.
    * An input image with byte ordering that does not match the native byte order of
      the system (e.g., big-endian values on a little-endian system) will be
      copied and byte-swapped prior to rotation.
    * An input image with integer data is cast to floats prior to passing to
      :func:`~skimage.transform.warp`. The output image can be re-cast using
      :meth:`numpy.ndarray.astype` if desired.
    * Does not let :func:`~skimage.transform.warp` handle clipping due to
      inconsistent handling across interpolation orders
    * Does not let :func:`~skimage.transform.warp` handle image NaNs because they
      are not handled properly for some interpolation orders
    * Does not pass NaN as ``missing`` to :func:`~skimage.transform.warp` due to
      inconsistent handling across interpolation orders
    """
    try:
        import skimage.transform
    except ImportError as e:
        raise ImportError("The scikit-image package is required to use this rotation method.") from e

    # Make the rotation matrix 3x3 to include translation of the image
    skmatrix = np.zeros((3, 3))
    skmatrix[:2, :2] = matrix
    skmatrix[2, 2] = 1.0
    skmatrix[:2, 2] = shift
    tform = skimage.transform.AffineTransform(skmatrix)

    if issubclass(image.dtype.type, numbers.Integral):
        warn_user("Integer input data has been cast to float64.")
        adjusted_image = image.astype(np.float64)
    else:
        adjusted_image = image.copy()

    # Scale image to range [0, 1]
    im_min = np.nanmin([missing, np.min(adjusted_image)])
    adjusted_image -= im_min
    adjusted_missing = missing - im_min
    im_max = np.nanmax([adjusted_missing, np.max(adjusted_image)])
    if im_max > 0:
        adjusted_image /= im_max
        adjusted_missing /= im_max

    # Swap the byte order if it is non-native (e.g., big-endian on a little-endian system)
    if adjusted_image.dtype.byteorder == ('>' if sys.byteorder == 'little' else '<'):
        adjusted_image = adjusted_image.view(adjusted_image.dtype.newbyteorder()).byteswap()

    # Be aware that even though mode is set to 'constant', when skimage 0.19 calls scipy,
    # it specifies the scipy mode to be 'grid-constant' rather than 'constant'
    rotated_image = skimage.transform.warp(adjusted_image, tform, order=order,
                                           mode='constant', cval=adjusted_missing, clip=clip)

    # Convert the image back to its original range
    if im_max > 0:
        rotated_image *= im_max
    rotated_image += im_min

    return rotated_image


@add_rotation_function("opencv", allowed_orders={0, 1, 3},
                       handles_clipping=False, handles_image_nans=False, handles_nan_missing=False)
def _rotation_cv2(image, matrix, shift, order, missing, clip):
    """
    * Rotates using |cv2_warpAffine|_ from `OpenCV <https://opencv.org>`__
    * The ``order`` parameter selects from the following interpolation algorithms:

      * 0: nearest-neighbor interpolation
      * 1: bilinear interpolation
      * 3: bicubic interpolation

    * An input image with byte ordering that does not match the native byte order of
      the system (e.g., big-endian values on a little-endian system) will be
      copied and byte-swapped prior to rotation.
    * An input image with integer data is cast to floats prior to passing to
      |cv2_warpAffine|_. The output image can be re-cast using
      :meth:`numpy.ndarray.astype` if desired.
    """
    try:
        import cv2
    except ImportError as e:
        raise ImportError("The opencv-python package is required to use this rotation method.") from e

    _CV_ORDER_FLAGS = {
        0: cv2.INTER_NEAREST,
        1: cv2.INTER_LINEAR,
        3: cv2.INTER_CUBIC,
    }
    order_to_use = _CV_ORDER_FLAGS[order]

    if issubclass(image.dtype.type, numbers.Integral):
        warn_user("Integer input data has been cast to float64.")
        adjusted_image = image.astype(np.float64)
    else:
        adjusted_image = image.copy()

    trans = np.concatenate([matrix, shift[:, np.newaxis]], axis=1)

    # Swap the byte order if it is non-native (e.g., big-endian on a little-endian system)
    if adjusted_image.dtype.byteorder == ('>' if sys.byteorder == 'little' else '<'):
        adjusted_image = adjusted_image.view(adjusted_image.dtype.newbyteorder()).byteswap()

    # missing must be a Python float, not a NumPy float
    missing = float(missing)

    return cv2.warpAffine(adjusted_image, trans, np.flip(adjusted_image.shape),
                          flags=order_to_use | cv2.WARP_INVERSE_MAP,
                          borderMode=cv2.BORDER_CONSTANT, borderValue=missing)


# Generate the string with allowable rotation-function names for use in docstrings
_rotation_function_names = ", ".join([f"``'{name}'``" for name in _rotation_registry])
# Insert into the docstring for affine_transform. We cannot use the add_common_docstring decorator
# due to what would be a circular loop in definitions.
affine_transform.__doc__ = affine_transform.__doc__.format(rotation_function_names=_rotation_function_names)



"""
============================================================================================
Understanding Observer Orientation Relative to a Simulated Volume with `astropy.coordinates`
============================================================================================

This example demonstrates how to use the `astropy.coordinates` framework, combined with the solar
coordinate frameworks provided by ``sunpy``, to find the intersection of a set of lines of sight
defined by an observer position with a simulated volume represented by a Cartesian box.
"""
# sphinx_gallery_thumbnail_number = 4

import itertools

import matplotlib.pyplot as plt
import numpy as np

import astropy.time
import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
import sunpy.sun.constants
from sunpy.coordinates import Heliocentric, Helioprojective

#############################################################################################
# First, we need to define the orientation of our box and the corresponding coordinate frame.
# We do this by specifying a coordinate in the Heliographic Stonyhurst frame (HGS) and then
# using that coordinate to define the coordinate frame of the simulation box in a
# Heliocentric Cartesian (HCC) frame. For more information on all of these frames, see
# `Thompson (2006) <https://ui.adsabs.harvard.edu/abs/2006A%26A...449..791T/abstract>`__.

hcc_orientation = SkyCoord(lon=0*u.deg, lat=20*u.deg,
                           radius=sunpy.sun.constants.radius,
                           frame='heliographic_stonyhurst')
date = astropy.time.Time('2020-01-01')
frame_hcc = Heliocentric(observer=hcc_orientation, obstime=date)

#############################################################################################
# Next, we need to define a coordinate frame for our synthetic spacecraft.
# This could be something like SDO or *Solar Orbiter*, but in this case we'll just choose an
# *interesting* orientation at 1 AU. As before, we first need to define the observer position
# in HGS which we will then use to define the Helioprojective frame of our observer.

observer = SkyCoord(lon=45*u.deg, lat=10*u.deg, radius=1*u.AU, frame='heliographic_stonyhurst')
frame_hpc = Helioprojective(observer=observer, obstime=date)

#############################################################################################
# Using our coordinate frame for synthetic observer, we can then create an empty map that
# represents our fake observation. We'll create a function to do this since we'll need to do
# it a few times. Note that we are using the center coordinate at the bottom boundary of our
# simulation box as the reference coordinate such that field-of-view of our synthetic
# simulation is centered on this point.

def make_fake_map(ref_coord):
    instrument_data = np.nan * np.ones((25, 25))
    instrument_header = sunpy.map.make_fitswcs_header(instrument_data,
                                                      ref_coord,
                                                      scale=u.Quantity([20, 20])*u.arcsec/u.pix)
    return sunpy.map.Map(instrument_data, instrument_header)
instrument_map = make_fake_map(hcc_orientation.transform_to(frame_hpc))

#############################################################################################
# Now that we have our map representing our fake observation, we can get the coordinates, in
# pixel space, of the center of each pixel. These will be our lines of sight.

map_indices = sunpy.map.all_pixel_indices_from_map(instrument_map).value.astype(int)
map_indices = map_indices.reshape((2, map_indices.shape[1]*map_indices.shape[2]))

#############################################################################################
# We can then use the WCS of the map to find the associate world coordinate for each pixel coordinate.
# Note that by default, the "z" or *distance* coordinate in the HPC frame is assumed to lie on the solar surface.
# As such, for each LOS, we will add a z coordinate that spans from 99% to 101% of the observer radius.
# This gives us a reasonable range of distances that will intersect our simulation box.

lines_of_sight = []
distance = np.linspace(0.99, 1.01, 10000)*observer.radius
for indices in map_indices.T:
    coord = instrument_map.wcs.pixel_to_world(*indices)
    lines_of_sight.append(SkyCoord(Tx=coord.Tx, Ty=coord.Ty, distance=distance, frame=coord.frame))

#############################################################################################
# We can do a simple visualization of all of our LOS on top of our synthetic map

fig = plt.figure()
ax = fig.add_subplot(projection=instrument_map)
instrument_map.plot(axes=ax, title=False)
instrument_map.draw_grid(axes=ax, color='k')
for los in lines_of_sight:
    ax.plot_coord(los[0], color='C0', marker='.', ls='', markersize=1,)

#############################################################################################
# The next step is to define our simulation box. We'll choose somewhat arbitrary dimensions
# for our box and then compute the offsets of each of the corners from the center of the box.

box_dimensions = u.Quantity([100,100,200])*u.Mm
corners = list(itertools.product(box_dimensions[0]/2*[-1,1],
                                 box_dimensions[1]/2*[-1,1],
                                 box_dimensions[2]/2*[-1,1]))

#############################################################################################
# We then define the edges of the box by considering all possible combinations of two corners
# and keeping only those combinations who vary along a single dimension.

edges = []
for possible_edges in itertools.combinations(corners,2):
    diff_edges = u.Quantity(possible_edges[0])-u.Quantity(possible_edges[1])
    if np.count_nonzero(diff_edges) == 1:
        edges.append(possible_edges)

#############################################################################################
# Finally, we can define the edge coordinates of the box by first creating a coordinate to
# represent the origin. This is easily computed from our point that defined the orientation
# since this is the point at which the box is tangent to the solar surface.

box_origin = hcc_orientation.transform_to(frame_hcc)
box_origin = SkyCoord(x=box_origin.x,
                      y=box_origin.y,
                      z=box_origin.z+box_dimensions[2]/2,
                      frame=box_origin.frame)

#############################################################################################
# Using that origin, we can compute the coordinates of all edges.

edge_coords = []
for edge in edges:
    edge_coords.append(SkyCoord(x=box_origin.x+u.Quantity([edge[0][0],edge[1][0]]),
                                y=box_origin.y+u.Quantity([edge[0][1],edge[1][1]]),
                                z=box_origin.z+u.Quantity([edge[0][2],edge[1][2]]),
                                frame=box_origin.frame))

#############################################################################################
# Let's overlay the simulation box on top of our fake map we created earlier to see if things
# look right.

fig = plt.figure()
ax = fig.add_subplot(projection=instrument_map)
instrument_map.plot(axes=ax, title=False)
instrument_map.draw_grid(axes=ax, color='k')
for edge in edge_coords:
    ax.plot_coord(edge, color='k', ls='-', marker='')
ax.plot_coord(box_origin, color='r', marker='x')
ax.plot_coord(hcc_orientation, color='b', marker='x')

#############################################################################################
# Let's combine all of these pieces by plotting the lines of sight and the simulation box on
# a single plot. We'll also overlay the pixel grid of our fake image.

fig = plt.figure()
ax = fig.add_subplot(projection=instrument_map)
instrument_map.plot(axes=ax, title=False)
instrument_map.draw_grid(axes=ax, color='k')
ax.plot_coord(box_origin, marker='x', color='r', ls='', label='Simulation Box Center')
ax.plot_coord(hcc_orientation, color='b', marker='x', ls='', label='Simulation Box Bottom')
for i,edge in enumerate(edge_coords):
    ax.plot_coord(edge, color='k', ls='-', label='Simulation Box' if i==0 else None)
# Plot the pixel center positions
for i,los in enumerate(lines_of_sight):
    ax.plot_coord(los[0], color='C0', marker='.', label='LOS' if i==0 else None, markersize=1)
# Plot the edges of the pixel grid
xpix_edges = np.array(range(int(instrument_map.dimensions.x.value)+1))-0.5
ypix_edges = np.array(range(int(instrument_map.dimensions.y.value)+1))-0.5
ax.vlines(x=xpix_edges,
          ymin=ypix_edges[0],
          ymax=ypix_edges[-1],
          color='k', ls='--', lw=.5, label='pixel grid')
ax.hlines(y=ypix_edges,
          xmin=xpix_edges[0],
          xmax=xpix_edges[-1],
          color='k', ls='--', lw=.5,)
ax.legend(loc=2, frameon=False)
ax.set_xlim(-10, 35)
ax.set_ylim(-10, 35)

#############################################################################################
# Finally, we have all of the pieces of information we need to understand whether a given LOS
# intersects the simulation box. First, we define a function that takes in the edge coordinates
# of our box and a LOS coordinate and returns to us a boolean mask of where that LOS coordinate
# falls in the box.

def is_coord_in_box(box_edges, coord):
    box_edges = SkyCoord(box_edges)
    coord_hcc = coord.transform_to(box_edges.frame)
    in_x = np.logical_and(coord_hcc.x<box_edges.x.max(), coord_hcc.x>box_edges.x.min())
    in_y = np.logical_and(coord_hcc.y<box_edges.y.max(), coord_hcc.y>box_edges.y.min())
    in_z = np.logical_and(coord_hcc.z<box_edges.z.max(), coord_hcc.z>box_edges.z.min())
    return np.all([in_x, in_y, in_z], axis=0)


#############################################################################################
# Next we define another map, using a different orientation from our observer, so we can look
# at the intersection of the box and the many LOS from a different viewing angle.

new_obs = SkyCoord(lon=25*u.deg, lat=0*u.deg, radius=1*u.AU, frame='heliographic_stonyhurst')
earth_map = make_fake_map(
    hcc_orientation.transform_to(Helioprojective(observer=new_obs, obstime=date))
)

#############################################################################################
# Finally, we'll create another visualization that combines the simulation box with the lines
# of sight and additional highlighting that shows *where* the LOS intersect the box.

fig = plt.figure()
ax = fig.add_subplot(projection=earth_map)
earth_map.plot(axes=ax, title=False)
earth_map.draw_grid(axes=ax, color='k')
for los in lines_of_sight:
    ax.plot_coord(los, color='C0', marker='', ls='-', alpha=0.2)
for los in lines_of_sight:
    inside_box = is_coord_in_box(edge_coords, los)
    if inside_box.any():
        ax.plot_coord(los[inside_box], color='C1', marker='', ls='-')
for edge in edge_coords:
    ax.plot_coord(edge, color='k', ls='-')
ax.plot_coord(box_origin, marker='x', color='r')
ax.plot_coord(hcc_orientation, color='b', marker='x', ls='')
ax.set_xlim(-10, 35)
ax.set_ylim(-10, 35)

plt.show()
